# certverifypkcs11 utility to parse, verify and view X.509v.3 Certificates from PKCS#11 and file
# 
# Copyright 2019 Vladimir Orlov 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

array set CRLReason {
0	unspecified
1	keyCompromise
2	cACompromise
3	affiliationChanged
4	superseded
5	cessationOfOperation
6	certificateHold
8	removeFromCRL
}

#Кодировка ascii enc
encoding dirs [file dirname [info script]]

switch -- $::tcl_platform(platform) {
    "windows"        { 
#	encoding system utf-8
	encoding system cp1251
    }
    "unix" - default { 
	encoding system utf-8
    }
}
#В текущей папке лежит пакет PKCS11
#lappend auto_path [file dirname [info script]]
#encoding dirs <каталог с кодировками /usr/share/tcl8.6/encoding>
global mydir
set mydir [file dirname [info script]]
#puts $mydir
set ::listtok {}
set ::listx509 {}
set ::pkcs11_module ""
set ::pkcs11id ""
set ::parsecert ""
set ::dercert ""
# pkcs11 or file
set ::tekcert ""
set ::sncert ""
set ::issuercert ""
set ::subjectcert ""
set ::notafter ""
set ::notbefore ""
set ::cn_issuer ""
set ::cn_subject ""

#package require Tk
package require msgcat
package require pki
#package require pki::pkcs11
#package require tls

    set ::pki::oids(1.2.643.100.1)  "OGRN"
    set ::pki::oids(1.2.643.100.5)  "OGRNIP"
    set ::pki::oids(1.2.643.3.131.1.1) "INN"
    set ::pki::oids(1.2.643.100.3) "SNILS"
#    set ::pki::oids(1.2.840.113549.1.9.2) "unstructuredName"
#Алгоритмы подписи
#    set ::pki::oids(1.2.643.2.2.19) "ГОСТ Р 34.10-2001"
    set ::pki::oids(1.2.643.2.2.3) "GOST R 34.10-2001 with GOST R 34.11-94"
    set ::pki::oids(1.2.643.2.2.19) "GOST R 34.10-2001"
    set ::pki::oids(1.2.643.7.1.1.1.1) "GOST R 34.10-2012-256"
    set ::pki::oids(1.2.643.7.1.1.1.2) "GOST R 34.10-2012-512"
    set ::pki::oids(1.2.643.7.1.1.3.2) "GOST R 34.10-2012-256 with GOSTR 34.11-2012-256"
    set ::pki::oids(1.2.643.7.1.1.3.3) "GOST R 34.10-2012-512 with GOSTR 34.11-2012-512"
    set ::pki::oids(1.2.643.100.113.1) "KC1 Class Sign Tool"
    set ::pki::oids(1.2.643.100.113.2) "KC2 Class Sign Tool"

array set ::payoid {
1.2.643.6.3.1.2.2 "МЭТС"
1.2.643.6.7 "B2B и B2G"
1.2.643.6.15 "Фабрикант"
1.2.643.6.14 "Центр реализации"
1.2.643.100.2.1 "Росреестр"
1.2.643.3.8.100.1.113 "Росреестр"
1.2.643.2.2.34.25 "Росреестр"
1.2.643.2.2.34.26 "Росреестр"
}

package require base64
#package require asn
package require http 
global formatCert
set formatCert 0
global loadfile
set ::chaincert ""
set ::listcert {}
set ::crlfile ""
set loadfile ""
global count
set count 0
set ::handle ""
global tempDir
set ::slotid_tek ""

proc loadtclpkcs11 {} {
  global tempDir
  global mydir
#  global tclpkcs11
  switch -- $::tcl_platform(platform) {
    "windows"        { 
	set tempDir $::env(TEMP) 
        set tclpkcs11 [zvfs::list */tclpkcs11.dll]
	set rumsg [file join $tempDir ru.msg]
	if {[file exists $rumsg]} {
	    file delete -force  $rumsg
	}
	::freewrap::unpack [file join $mydir "ru.msg"] $tempDir

if {0} {
	set tempDir $::env(TEMP) 
	set tclpkcs11 [file join $tempDir tclpkcs11.dll]
	if {[file exists $tclpkcs11]} {
	    file delete -force  $tclpkcs11
	}
	::freewrap::unpack [file join $mydir "tclpkcs11.dll"] $tempDir
}
    }
    "unix" - default { 
        set tclpkcs11 [zvfs::list */tclpkcs11.p11]
#        set tclpkcs11 "./tclpkcs11.p11"
        set rumsg [zvfs::list */ru.msg]
#        set rumsg "./ru.msg"
	set tempDir "[file dirname $rumsg]"
	
if {0} {
	set tempDir "/tmp" 
	set rumsg [file join $tempDir ru.msg]
	if {[file exists $rumsg]} {
	    file delete -force  $rumsg
	}
	::freewrap::unpack [file join $mydir "ru.msg"] $tempDir
}
    }
  }
  load $tclpkcs11 Tclpkcs11
#catch {  file delete -force $tclpkcs11}
}
loadtclpkcs11
    
::msgcat::mclocale en

set lang "ru"
set lp [llength $argv]
if { $lp == 2 } {
#    usage 1

    set lang [lindex $argv 0]
    if { ($lang != "ru") && ($lang != "en") } {
	puts "BAD (en/ru) language=$lang"
	exit
    }
# 1 - language en/ru    
#    puts "LANG=$lang"
    ::msgcat::mclocale $lang
    if {[llength $argv] == 2} {
	set loadfile [lindex $argv 1]
	puts "File and language from command line=$loadfile"
#    fileWithCert $w $f
    }
} elseif { $lp == 1} {
	set loadfile [lindex $argv 0]
	puts "File and language from command line=$loadfile 1"
}



namespace import ::msgcat::mc

if {0} {
::msgcat::mcmset ru {
"Cancel" "Выход"
"Details" "Подробности"
"Identity" "Подлинность"
"Verified by" "Проверена в"
"Expires" "Истекает"
"View certificate" "Просмотр сертификата"
"Load certificate" "Выбор сертификата"
"Subject" "Владелец сертификата"
"Issuer" "Издатель сертификата"
"C" "C (Страна)"
"GN" "GN (Имя, отчество)"
"L" "L (Местность)"
"Issued certificate" "Выданный сертификат"
"Version" "Версия"
"Serial Number" "Серийный номер"
"Not Valid Before" "Не действителен до"
"Not Valid After" "Не действителен после" \
"Certificate Fingerprints" "Отпечатки сертификата"
}
}

array set dn_fields {
        C "Country" ST "State" L "Locality" STREET "Adress" TITLE "Title"
        O "Organization" OU "Organizational Unit"
        CN "Common Name" SN "Surname" GN "Given Name" INN "INN" OGRN "OGRN" OGRNIP "OGRNIP" SNILS "SNILS" EMAIL "Email Address"
    }

::msgcat::mclocale $lang
#puts "LANG1=$lang"
::msgcat::mcpreferences

#::msgcat::mcload ru
::msgcat::mcload $tempDir
##::msgcat::mcload $mydir
##::msgcat::mcload [file join [file dirname [info script]] $tempDir]
#puts [file join [file dirname [info script]] .]
font configure TkDefaultFont -size 10
font configure TkFixedFont -size 10
option add *Dialog.msg.wrapLength 6i
option add *Dialog.dtl.wrapLength 6i
global home
set home ""
global typesys
set typesys [tk windowingsystem]
#set typesys $::tcl_platform(platform)

#puts "TYPESYS=$typesys"
#switch -- $::tcl_platform(platform) {}
switch $typesys {
    win32        { 
	set home $::env(USERPROFILE) 
    }
    x11 { 
	set home $::env(HOME) 
	catch {tk_getOpenFile foo bar}
	set ::tk::dialog::file::showHiddenVar 0
	set ::tk::dialog::file::showHiddenBtn 1
    }
    classic - aqua { 
	set home $::env(HOME) 
	encoding dirs $mydir
    }
}
image create photo butborder -data {
iVBORw0KGgoAAAANSUhEUgAAAEAAAAAQCAYAAACm53kpAAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAA7EAAAOxAGVKw4bAAAAB3RJTUUH4wIXETEq6wj46QAAAB1p
VFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoIEdJTVBkLmUHAAAAlElEQVRYw+3WsQ3CMBAF0MuHzsahsSWEBC0LpmMYmmwQyaKIIjEFkVAGsGfIXap0bHD3N/jvX3HN
PH/HZflNIgJSEhFZnfO3lJKnnIcnKU3OQwdNy/+5BKgtv8cADMAADMAADMAADEAvgIiwttK11v0VZrTt+aENIMZIRETeny5HoLn2/etdSvkAOGgAYOY1hHAHyG3Aui+d
rR5/OQAAAABJRU5ErkJggg==
}
image create photo butborder_ORIG -data {
iVBORw0KGgoAAAANSUhEUgAAAIcAAAAiCAYAAAB1NspXAAAACXBIWXMAAA7EAAAO
xAGVKw4bAAAAB3RJTUUH4wIMDAsSdUV+swAAAB1pVFh0Q29tbWVudAAAAAAAQ3Jl
YXRlZCB3aXRoIEdJTVBkLmUHAAAA80lEQVR42u3VsUoDURRF0XMf3sh8QGxiEeyC
TaJ1wCJBvzWdRdAwKYVUQhpJaYwz3euneMW1Eiy0F2avTzhsOBYRatsmDoc3lVKE
fnN3TSbXGo0u7axtmzidPjSf31lKSe7OQj23272EmcLq+jkWi3tjEvy03W4icSX4
TSlFiRnwF+IAcYA4QBwgDhAHiAPEAeIAcYA4AOIAcYA4QBwgDhAHiAPEAeIAcQDE
AeIAcYA4QBz4J3G4u3LOK6bAt5zzyt1lTfMZ+/2rhsMLTac3Nhics07PrdePMZvd
yiJCdf0Ux+O7uq5jmZ6rqkrj8ZWWywf7AmxXRGaRW9gDAAAAAElFTkSuQmCC
}
image create photo openfile -data {
iVBORw0KGgoAAAANSUhEUgAAABgAAAAVCAYAAABc6S4mAAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH4wIOCzQr3dgCiQAAApRJ
REFUOMullLFvFUcQxn8zu3fYeo7lR0CCkAbJDaKjoQju3JCOhp6SFlEh0SCkSBT8CbShiVCaVDSIBkXpiCuiUFAgWWDA2Jh3tztDscfzPb9neIhp9lY3+30z8327wjfE
P3/dYPfun8NqQf4Otax+LT+9twcC8Hj9rKw93PB5SB79cubKibXle6d+Hc5OaMuiA+XpnRdEgLWHG/54/ezV1NhNoDkUfYSHYTjx0/oy7qUeqWT821sHN6yHEMftNHbz
/J3TJ/NuLgcRHEcqQTOwAIhCBGusVFnrBL9hOIpi0wRAk3czNrJxi0Cp9IeAGZCMWClERQWCB7I6wYQsGdAxCTJNMA4d6Lh1rRWtQAGOKMEDAaBTLFr5DB6g2ichAQni
87vnf1xZ2vR/7xNL9YaNQL3M10YGrU6MQavpoqwF3MiNgyihFogQF38ePllZv7Z64bJ+Rvi+kJqXv//GyrmjpcNwbLDK8HKHrL31W2LyzGj71rTIzkcE7RL7BPoF0Fnr
AirMctGHov53EdjUjHsEO+ACIoeZ67AHoRuBg/jUPY37SVsg3dZDRzRHuIPkzvepQMYDBM428Gqf7zPRV8U2kHSgm1hu/GQHOx1BffDKzeGe/ncqGBMEyYAdxF5hGmeI
PK9NDbWeXfes14HvQn6LdltTnbMDesBAThDKiCz0RyR7kLYgRwjag51zRBnIBm5TDux2DfARXLtHSmcaeZY7p3RI2vsHsQB04qaS7Kkli2A5YW5YNshgYmgoGql2h6MS
VBCRcTHDi8fZ/GOT3Ph/cfTszYOXr29farf2SCOjfZtpPzjtZqZpGvKe07yD1Bi0hgwq6iNGtSiE5UC9FKmWKuqjEAeKLipRor3/v70/2s7XPwFC7BN5lLYzNgAAAABJ
RU5ErkJggg==
}
image create photo svitok_40x16 -data {
iVBORw0KGgoAAAANSUhEUgAAACgAAAATCAYAAADiQ08DAAAABGdBTUEAAK/INwWK6QAAABl0RVh0U29mdHdhcmUAQWRvYmUgSW1hZ2VSZWFkeXHJZTwAAAm0SURBVHja
Yvz//z8DtcC1z3+5GRn+8wCZTP9//xFn+f1Lg/nFM6U/4hI//3PzCDEyMMgzMjIIAuXZgFiCiYFBhOH/fy4gmxkox/aPkXG7Kg+LP7KZAAHEiOzAK5/+ggxX+MfAwA6k
/wM1CTAxMkgBaZDhIMvFgLQYAyMDD9BgTgaIBYJAMRagOexANg/QAWw/371leD9jGhPDsSOMQpycDJz8fAzvgsIZuF3dGFiYgM5iBJkOoRihbCagRexsrAx/GZgS5biY
F8DcBBBAcAfaNMwXUJESrbHRUY40UZER42BlZgErABFQNX///mX4+P49w88f3xkERYQZONg5Gb58/szwHyjOzMLMwMvLC7bx1YvnQCezMzzt7WEQ3bedQRDoSCE+PoY3
YTEMMskJDExANSAMCjYIzQjUBqSBjvzxj+ElLwujBMyBAAEEdqB71wpdAR7u+V+/fTN+9vIVg7QwH0NluAeDhAAP2IeMIAik79y4wcDLx8uwfdMGBlcPTwZhYRGGyxcv
AD3wj4Gbm5uBE+iQN69eAi1iYpCVkWW4dPUaw+djJxmM9mxjEGVjY+AQF2cQWL6CgUdYmAHi7/8M/0Ae//ef4TcwDH79/c/wh5HpsiQHkz4kXBkYAAKI+RSPlriitMRO
bk4OHZAFbOxsDI9fvGXYfPwCg7OhJoMUP9BiVmYw5gJGwXugA+5cv8agq6PFcPb4cYaPb98wiAgKMNwCigkL8DPwAR2qrqbGIAQU+//nN4OyjQ3DDWCwcFy5yMALdOTH
z98YmMwsGD79/svw/udfhjdA/OrHX4bXP34zvPnxh+EbA/MFKQ6mJbAQBAggZhXXUJ9vP3/KvP/05dinL1/3//r9ZxfQkXtYWNm2P//wZZ+9toIjHzsLExszMwMrMxPD
owf3GYyNjBgkgKHBD4w2fUMDhl+/fjGwsrIx2Ds6MXz59pXh5+8/DHfvP2D4/P0Hg4ScPMNXeVWGVydPMYh+eMvA/PoVw3OvIIb3v/4xfP79j+H7n38Mv4AhCAw8cJCx
sLBcBjpwJcyBAAHEiC8Xv/j6k/H73/+LJLnYYpgYGRhgyfElMBT379/PoKGpzcDMysrw+s0bhnv37jH8B0atvIoaw+vXbxi+//zFIKeuwSAoKs7w8dtPhnNr1zHYzZvA
ICciwvAiOYfhn28Ahn1//gGTChdXlYkASztMDCCAWGCMS49esrGyMAsBE63wr79/dT58/W734t1Hqw9ffqg7GWsyiHBzMnz9/p3h4cNHDCdPnGC4du0Kw8dffxnYuXkY
OLl5GeQNzRh+/vrDIKKoxMAm/onh188fDEKiYmCzOTkZGVh09BhecnAziH77xiC4bB7DWxd3ht/AKP/+6zfDh68/GN5//c4gKsB7Rpmbex6yowECiPH333+M775845m8
62wJHydbgoGsqJi8KD/HJ6CvH756z3Dr2WsGLm4uBm9LA4b79+8xXLtyGWixBAMXvwCDoIQEMM1ygnMgPgCKpev3HjK87elgsLxwlEFeToRBsJSXgUOZjYGd8TMwi7H9
uMJZXPRHwHPeZwudX/xMTMoGV6/eBekFCCBQCDL79azsf/fpc/I3oO84WZkY3AzUGUIdTBn+AX34n52d4cGbjwxPPn1n4BKXYTAUkWBgZ/4BKXaAxaXKzykM3P8eMlzg
6gGL/f71k+Ex0CPcwPTJycPHcPPKJQZGNnYGZn5RhvOm9gxaD44x8Hp+ZeBX+Mnw/fOv/09fvGcUkhLg1PtfNe3LsYqzd5iZ44FJxQRolAXIPIAAYrFrWRLy+efvZEZg
sQdK6F++f2NYdegcw6svPxh8bE0YmFjYwMXAhy/fGEQF+RhYgIWX7K+1DGK/DjI8Y/djEPu5BhhCzAy/2f8B1f1jePfuPcPlCxcY2Hn5GD5//cbAADSTW1Ke4d/HrwwC
khI/ODO1fgvbPuJ99+ADw5s2bkYmFrHXHxM/snAYswly8TOf5OPjZvz89evLo6qqm4HuWw4QQCx//v2PYwE6Dph7GFiACZ4JmFu/fv7IcODcNQYzbTUGLg42oMNZGH79
/g1JE4z/GMR/7mL4+PQiA9OnIwy/FQQZWDk4GbQ+lTN8+8fNIPD9P4OY5icGdi52Bh62nwxC7F/+czP//sLJ+P3pP+kvl1mY30n+/v3X+v35//+u3fnAzM3OLirUzsUg
Oh1YUEv/Z3z+5jXD1x8/OYHuOP3q3buVAAFYqmMbhGEgCqDfDi5AQsgpLKEUiAZYg1mYg4oFGIcNWCMSVSQaRyQ4d4nNYSiu//p3ujeLKW7lnWcxjISUi8bIhP7VoW6e
qFyJTsLRFBEooOIbCllpubYy6fdPE8OpexZ1syoEyJShVJr/YkKIwQEa+zTqfmBSix3Oc2NcYD42kW1bL62VloaJLg/vryfv399CPgKIBZhrH4A1w7I1KDSBBTIo4X8H
OuzNl+9A+i/DP0amf8DMxMT36SbD/Te/GYCxzsDy5yOw1gBG7Z8/wLqUiYGZjR2SrIEFNDPQ4f+Boc3Oy8bw4+tPYEyzMPDz8DIyvGHh5FBlZxASYVfj4P6dyAHUIZX/
k4XXmHUjAyujptkSrh7mJK7fMPcABBALGwvLpl9//nog8jXI5//BdevNR88Zvv94yMABdLCFjhrjh+//GO7/jmD4yRnC8BVYzv0DqpMQ4GMQ/PLhLd+TBwcY7txa+2/z
ukuRvKrX3vdeYfj94xcPMF2KfL3Iafv9IZPWzTt/rH//+2tj0w4MBCbmKO0SlsVXe/7sk9RntwL6z/bfv/9vgBkepWAGCCBgoDIvYWFmyvj9558epGXxDxjjrAxsHBwM
L959BqZLNgY5WVGG15+/MoLKLKBnGIT5uBk0xYSeyUmLb+fn4lwELBoOIRu698MVhnPJIBbbZyABwvfh/mdgkHt+4/N1cT1uLn5plm3alYxNHNxMJcBQ5vnz5/+S88ks
f5DNAgggcE1i27rM9cuPn5v+/f/PAaoqQK0WUI0BqvSBdRyDmZYSg6QQH4O8qNB3IN4jyMM1Daj3ANBhP8hpN8om/hJikfnyVExFkIODj43h26cfDO/ufvr5bq7YPsYP
bHvFeXhmih048AWkFiCA4FWdVfMSrx+//0z+9++fEiyMWYENBGcjzd8mqnKvFCVE2oDpdS/QUTep0bjVmvzD6OfHH4cYWRi4Xx1gY/i8m5eBi539Ax8bW64QD89ykd27
/4LUAQQQSl1s07pU6e/ffynAOtESmFm4zDXk71pqKe1WlRFfZMTPDNZwUFOTgdrgx79/LsCAsf/DyMjz+///Iz9//lwX+egR2GEAAQYAon2R4+gZwDgAAAAASUVORK5C
YII=
}

ttk::style element create RoundedButton image \
        {butborder} \
        -border 3 -sticky nsew 
ttk::style layout My.TButton {
  RoundedButton -sticky nswe -border 1 -children {
    Button.focus -sticky nswe -children {
      Button.padding -sticky nswe -children {
        Button.label -sticky nswe 
        }}}}
ttk::style map My.TButton -background [list disabled #ffffff pressed #ffffff  active #93cee9] 
# -foreground [list disabled #a3a3a3] -relief [list {pressed !disabled} flat]
ttk::style configure My.TButton -padding 2

image create photo up -data {
iVBORw0KGgoAAAANSUhEUgAAAAYAAAAICAIAAABVpBlvAAAACXBIWXMAAA7EAAAO
xAGVKw4bAAAAB3RJTUUH4wIKDRkw/Hh6bwAAABVJREFUCNdj/P//PwMqYGLAAPQX
AgDFpQMNg52XXAAAAABJRU5ErkJggg==
}

image create photo creator_small -data {
R0lGODlhPABOAOf/ABQUHBwTExYVGBgUHhkVEh8UEB0WGygVFxkaIRkbGRwbFSAaFSQYHiQZFxgcHhwbHh8bGiQdFCgcFB8gJx8hHyUgGiMgIx0iIyofGyQhICkgISMj
HDIgHjcgGyskGTAjGjEiJy0kJC8kIDUjHColKSomJSwmICUoJzsoGEEnHjspHUAnIzItGDgqJzsqIzYsHi4tMTEtLDQtJzUsMi8vKDguLy8yL08qIkssJEUtL0MvJEcu
JEIvKTkyLEguKT4xKjY0KDczMjU1LTM1OlExKE4zKT43MTs5LUo2JTc7Ljo6MjU6QT86Kkc3MUs3K1szJE81Pl8yLkE5RFQ3Iz87OjU9STs9OlM3M0M9Jz48QDk+QFo3
Jlw2Llo2NUY8P1g5L2E2OD1BSGE9NmM9MWA9PFpANGc8PWw9N0VHSlFERUlHRUNIUGdBME1GTnE/MlRHQlRGT0tKQ2NFMFxGQXNBRGxFPGpFQ2JJPHJGMXpCR09OUVJP
TYZHOXpKQlpSV3ZMQnlNPXJOTXBQRIFNP2RTYV1XVntROlZaV19YUXZTP2xVT1hZY1laXY5TWIlWUYtXSYhbQG9gYH9eUGpjX2NlYIhcTWtkWHZhWZFbSHFhbGBna4Rg
TGZmbIZfXGtlcZViVpRnWmxybn1ubJBpaJVqUXZve5JsXnJ0d4tvYXJ0fpVuWnt0bmx4gX97jKR1ZK10X595WKF5X5p7bXyChJ97ZaF6boWBfJx7d32Di3iEk5CAfLiD
eqyJbKmIhayKcrCIe5qOiYmSoa2Lfa+NaI2Sm46TlqWOiKiPfrmTdrKWj8WShLeWhr2UibyYgZqfqJyho5mhsaefnL+dj7Khlrmgj7qfmMaio6WstMGmn8elmcKnmM6k
l6qvsbKuqL6tpKmywciun8+toMqupsyxtcW4r7S8zNq1o9i2qda2sNK4qdG5sNe5pb6+wLvAyb3E1N/BrN/DtOPDvd7GvcPM28zMzM7P2crT5OzQx9DZ4dHc7dfk9vrf
0+Pq6eDt/ebz+////yH+EUNyZWF0ZWQgd2l0aCBHSU1QACH5BAEKAP8ALAAAAAA8AE4AAAj+ABMkeGBAAwYNGg5oCFEhw4YEDhxYsCBAAIUYMnrIMFGiB5AeIIGIBCKk
JJAjST4eYXLkCBAlQjYEMFDRQQIBDxJMzLAQoQYRIkqUoCAgwQaiFmnQiME0RpAeT0H2cFmyJUqVTE4eUQJkAwECRQVCTECBZwiEGESYCGEiwwMKZB9UpEBDho0YNIRI
lWoEphIlLZMItlqyZA8PCyoKzGkzAQSfbEWcDWHhAQQIFhDMZWqjZOe9RvRWdSlYsJCWhYUgBrt4oMDHFdZKRuj2gYMHJB4AGHCBRA3PnT2fTi1EpNWWI0/GXCBWbETH
FSqUCIGhAu0Hby2UsDBgAA0rVNT+xFFD3ooVJeb/CvkbePDxkRUgKBZr2/GGhhkyrA1R+ygFEmEwEsopnJwSSiicHFJIIQr6cUgcSfyVBEruWQXfV85h54ABD2zg4Ych
hLDdBRCccMgsxBRTDCuzcKLJKZowwsginHDCyCHnlaREacW9B4QMGmDYGgIIYEdWWfllUAIJF5wQiopQ4jKLJlRqssiVVCIYB3HoWTGaS0cUt8ECCogFAUQ1CbRTBg4g
MMEJszyjopyz4AKjJpQwYuUilIQyZRxKBCeEeVZsBRhyQtAwZpmvJaDZfBmciV1OhTxjKTHO4CIlJ6mkwkqVp5wyyyyhUGJFZyTtOCFMJ8GUKAT+BZzp2kDYFUkBBRbQ
moAN3FjqDKbESFlnnamcwsopxcLICBqpaVUcSIX9WMECZJZ5mQURIXACCdyS8GYCz3DTqzPkEhNsMcJy4omLVObJSI6F7XiEUj1EK0ME1SqgLwQTXOAvriW0VRkavT5z
zcHlFmNusLh0qgknlByihx7nCQqYEjIoJQNJxXmgr7WX9dvktjCcEJEAlhosbsHPkItuKlSuy8kie6iRhRdG3FWCaErUgFfGStGAQQMNVBBBdBVYQMEJMMDgFAkDWdEN
MSqLm3KwmqzBiaYw6vHGDzzw0EQTR2ikWgkYxXCXDTSYsEEHI2DwgQcfzH0BAhbYIJT+iBc4oEbLgFuq8KeL4JLiKX54UQPYYfOgww9GaKRRCQycIIMMJaBa1wsvqMD5
CyxYMMEEJCzJ7USHkNuyig1TiQu5pxQCRQ4+7ICEDjzQHrYOTfzgAkglXBYw2zS0ZAQWWLDUgxGiTzQRCRpYoAEjdK5iCyWr4FKMwcREMkcZZIgxRhFF7KBD7ebr4ETY
LnDQQgsVbCAUUxlEFwG+C0SAAd7OT4SdAWEoxixsYQtgFDAachJFIBJhCEAA4g91GAMRnFAEHeggfRZcwQpawIG0LEBJQpkWvgpAQglIgEhFmhR2JmAAZ7CjG92Ihgxh
CIxj/IIWqogFKCpRiU1Iggz+VyjfDpxwux1osH0HiZRQTGC0+y2AhBE4IQofwID+BQGG3SBHN6YxDXV4AxzpAAc1lnEMX8CCFKSoBAS5wIUobGEMW0ifC0bwgQZsoAEW
IEGk8reArxSgASd8AAIGAAABbGgCFKBEFsnBSG84UhvpCKMYj0ELUmxiE2o8wxnYcIYo1IENFdTBHD+AgQg0AAKRmtYCAkCAAASAaESSSwofYAEY2GKRjCQHNr4YyUhS
wxeq+IMchtmHYvahDmeogyC+IMoV0LGDp7yMKltJgD82YIoAwI4FsuAHXDayGmCMJDi0QY1jqCIRbBgDG/rgiEGocQx1qEMZ1qcCFXwAmpf+uQy1XElCWKIQAAgwQN6A
QQ51ZNGR0wCnOKmhjWWgQhJlKMITnsAGTJzRDTcoAhLWx4M5jmCO0clAH1cZgAK4kmgVSaEBDMAAIXgDG/JwpCPFUQ1xgGOc1KCGJGQhCydMYQtthMQwSPGEHaggBSlQ
Afta4AJSaiA+X6kmCV+JgYpkM6ArjYM2dhEPLzoyHTUdZzNoQQthiKMXlQCEHLjghlgggxRsQEEH5lo3FYhyjhwQgQeiGtVXolQAhSQIh1ogjVecw6beUEc6tGFTcAy1
FssQRi1AgQc24IEUzViGK86Ag7lyAJAjSIFH87rXvpZ0ARVoAGABkE2CNIADynj+xTbQIQ55qGOc6rjtWGlhRlVAwhCVcCszmKGKM9xgrh3gAAdGENpniqACJm1lAJ64
AAxwYDesFcBKD3CAXexiG9LQRkyxAUnFasMXvIAFJipBiljE4hfgFYYp3LCFFCR3BXOcK3NHIAIOsNIArixpAzAggQeARbsrZcABXKEMZTiiGjGdBiThIQ90LGMZlSSF
K14RC2ZsIxvSMMUZJGjfFaRgBR1wJh1dgIECrNSVCxhwKbGzmwGs1AAHAMU2PvEJadhWHJGUR4WlsYxfuNcVpMDEL7JhjnDUwg1nIMIOkEpl0Ta1BR/oJ9Hy14AFFDgz
3bkxA/6wDVfE4hjwSHP+muNBWxAzgxc6xAQmmiGNX0h2EH3gAg5wUGUc4NcFLRBBlytQHaNJIIp5PMEELMBSntwBvNKQBYWFbFt1ZOPSzPiFevkw5z5EQRby/QMRvrCD
Hex5zytAgQt+8NkIaCADHqiABzzwgg88byIaIAEISnAHIp8jHJOWBzwUCw6aMsMXrhjEIEzBC1LU4hagEMRRi+CDU/N5BSr4nQcAiQFae6DbHuBABiiwnW6BYAY5GO45
4nEPSsMjHukQB6Zr8QlHCMMXvuCpJARRhnlecAdFIAIOdjBH321bAuCuG61doCQS5JFb567BLbYB7ElPWh3iEEedQeEKMh6Dp7JAxR3+BHGHOdwOB0QQogtc0FEJDHjW
Clf4CfYWAojXoAaKiMc7gi1kjIsDG8swxS+kkYxeGOMYqECFIu6w9Dc4oXZFGLgPOvqDFmDg6hGY9f2+bYKZLylEJJhBDWbwA13Eu92U9rk2jiEMbRjj7Ue/hSkCcQem
f+8K1S711HlQ9aFdXW6z9oBaSjBzboUgBmKfARSacIlq3CPNPbepNpKRDGwYQxfGQIUpRhEIMZThCt+bQxOCKHDc8Z2DDfjAC1yQVBTUUwVCWVJTxE4FL3hBEbVVMzwk
+fNkVCPpx1gGM3rxCTdwgQg+yMHY8B51Uf6A1cql4wdG0IG51W1JM5gBU7L+XwMlNGEOl2D37tMBj1+DA+PTuIQstAEOIvfiEX0YMbVX8IPlO+5xPbB6WmZ9EKRVAAYk
0DQzkBFBUANG8H2o0G7lR360JQ65NQ2SIA3yFlmqYAhuMAZj8AVFgGI+4AROYEEqADkt4HJDQzRbFh0woH1OEQPL4wW1dwW3oIDpEA7kF0bxZgzCkA00hWGkMAhnwEZE
IGUAdwVNEDbPV3UuBwHUElX55DRqAxVBYARUAB6UMA0yGA7Adg6RlA3VUA2XVg33RgufkGdPEITl4wRl0Ds/IASRY3UNsQEKQABicSsx4DQ2EARBQAW1ZwWnwA7YEA9p
9mvhcA5aeFjogA7+4ZANmcYLrvAJgzAGxydwReCBfBcaR1B1IQWHzUGHToOHeXgzcUAP/KAOgLh7WEiIWIiFh3gOYMgLqgB/UVCGAucDPmBBP2CARmAEL4AB+VECcSiH
CXACJ1CHMDAEQ5AF4WEF3OAP/UAPpAgP63CKiSgN1KiD6FANRkYKj9BJbHR8OFA7IfgUiyMDDTERArEBJ2ADVjAENmCMQyAFWUAeocCM+lAPh3iIqRgORHZh1EhktWAK
oOCDkPgFszh1HgEEuWgEMoABZQEXG0ADd2gFNDAEVpAFUiAFthcG/NAP+tAP+TAO4oAOgjiNF7YM1ChZkzWGf5BOX4B3PrACPND+A34ROeSYH7eyNDZwhzGQBTyZBWEQ
BpxAj/owlO1QDdggb5cGYiZJjZEFTJUwCBCEgQTpkpCTh0qQiz3QFuMmEOloAzfXNGtQCq0ADfjgkfnQkf1gD7pQDdUIYvu4DM3QDMjglJUQQXAUhLTIAyyxHlepEeN2
K2TRlTBgBdtzDd/gDvPgDh5pD0PZD/3wDJewDNTQj8uAb8LAC/nmCqqACfHHBV9AQdQ2dVX5F0YQBPBjk0vzHTBgA0qAmNdQDvPwDeWQlvZwlo5pD4gQCMIgmZHFC8Pg
C7GgCpOlCpUQfxrYBHg5dbk4A30RJjXAiw5nAReQkzlpBe1QDu1gD/b+AA3foA/24A62yZHOcAWCUAvCIAy0kF6xYAo85AhqFEEaWARdkJdHkJDscQRAkkfO05VagAb1
8A3tkA/lAA3QoJ3giZZpSQhkcAadYAqqYAqbYAjx9AcPVAdiIAYaGER49wMt4Sp/8RT8ITq3IYxZoAdo4A7f8A3bSS7z8J21iZb64AwrAAZ2IAiS8Ad9wAZicAZ/EJUX
+gUt6QP29QE04KEf2gMi8B8lcwLGqAVhsAQo2g7uAA3OcA2IiZhn2ZjuAAVdIAbxNAZiwAVkUAeb8Afi8wUXWgY7QAajIHKKEAlvYChCEAQ/EAIwsARDoAVa0KR46g7g
+Q3k8g2x6Q7+5cCYjWkPkXAFXKBOkNgFXVAHiVAHXQCmX3AHl+AN8SAO1lANlHcMktAEf0EFNRAC/akFS3CqWaCnVuAOtUmgzqCiglqo+mCb3zAJV/AFiwqJXOClyEQG
XHAFqKAOtjUO1mANySBZpHAGoGoENwcCVqAHayAFYcCTqsqq20mg4BkM5aCiZ5kP0LAIVxBEX4CBbSQ+8SQGnYAN7aap1dALyXALnVAHX5ADoVqAIKAGa1AFVXCq/Aql
+YCiB4MP5ZALslmwucAKcBCuQDoGUcBGaCo+ydBuGIcN1jAKdqALi4CRU0gFxggDLUACWpCv/XqqQ1AOAoqighoMwZCiwQD+DQcbDAmrqIsaBWAApGJwB+cgZL3XC53Q
B5fQDhmLjFQQsm0gBSIAAkugryOLp/Pwr/Ygm+WgstAwoPMQDKzACl5wq+MqBlGAoVsgCeFwD2H0rpIQCHZABrowD4swrVOYBYvgCcFgAEirr/t6qnWbDy36tFOqKVM7
oGsAB1BwBY7KBemEgWOQCO8gttiADbdABnZwfFcADPWwBnrgBT3JCJ4ADQNQAklLt0mbBUlrD/Mwq4cZDKlgONuKbmDgqF2wumOQTGMgCOmwD+qADZwqCILgqb8aDfWw
CGjQkyXqB9CAAGrQBmFAt1Xwk2GwBgeaD6V7usQADVIQuF0guI7+agY7uknLQLuLWw2jIAio0AlmSwbR0A6cAAcTowZ6wAh6gAsX0AaM4Af5qq9hgAZrcKL54JHl4A6t
0AqpUAypAAWBSwYETMCa5KXUsA/woA1d2At/EAidcLZiQL7skAp+sAc1gwZ6MASpMAF6UAgyIr9VoKc/qZhpWbWtgCzFQAhQAAbhc6F2YAd1EAWmcA8KXE7GwLN1wAVm
ULNXQMGccME10wZo8Ls2QAkLMgmTUAh64JNZgAbz0A/+oA/f4L+hUgpQUMDxZAZmYAeB0Avy8HNdaJS2uwy1IAmS0JJkAAzs4AlCrAZpAAeWWwNpgAiIsMSFEAnsuwRa
kL+OGQz+pVAsq5AJBSzD8VQHknAP6EANY0zG1NCpsmAKYvDDpfAMQYzBNZMGakAFPfAGSlwIlGAJk4DEe4AGfpwPweAJpVAKkbAKokAGdxDD8SQJ7Fa7tjvGxgByqCAI
8vQFZJAGq7DEmFwzNlMDlKDEiGAJokwJzMwI+sCM7pALq5wJomA9osAGMYy78LAPioxxjdwLkSwJFloG41oHTYAIlDDMe+AFm1wDh6DE8KzEkTDKUuy8qxwJopDPtqAL
l+DFssvNtrWp50l5RScLmyAG1fsFZSAGOeAH6TzME6MFnRwJheAH8zwJkTDPlOAP/rCdpUAI1awLurDP/CwI4QDQGPeirqAACrXADJwayXUQruFDBj7wBg+tzllQA4Gw
03YwB60sz5PAjFFbCtS8CtZTQMCgC+KAdoeIDQDZCY3gCJ3QC7WQDL/wB+QMpF2QA548zIiAwYrTCY5wBmc7B6JwzErMjN8gzZlQCiNN0sDgDfdwD7mlDtZwC8XkCI6Q
B3RAB6CADr/wiOKD0OccCRhcCJjsB1LQAoEQfxNMBoGgCBg9CQEBADs=
}

image create photo ::img::ru_32x32  -data {
iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAACgElEQVR4nO1XMW7bQBCcuVsxlm0IYBpLfkJe4FRJmSrII5zSvYtU+UYekL+4SOHKQMrAAgIVMWyAoGTe
bgqS4lGmlciilSZbcA+n5c1wNLdHEg+DHXN9hT0Gtpq7anoDjeZMjo6O6L0nyT8ReWrEa7XIzOdz4+XlJff39ydJkgycc3zkxj7CAMDKwGKxuM2y7Jfc3Nx8Go/Hn5Mk
6RlvDRMzZFmWX11dvZXr6+s3JycnEJGdEQCAJEn2Li4uXgtJTxKVB3YWJDGbzcTtFLUj/hP45wRkPjdkmZr34UkLWFefWxO11/NckedmcnoKnJ0BIoRzZUGd6+J4g5gB
IZRZdT2B+LfV+hCIoiAEIFQdVV0LbDWrNsCb7lizBrgmokqEAEpXkWqpQj1WbRbru13IKngdNfBz9Kd4zVb/jRXYVWNsKVBL3WW+PiNW28WTXY7edJttGlIe02Zm2jwx
sfXbwDreJY5BUJh8kY/2Pjkz7xzAqAH0dUIaYLBGYlNADXMNONccMnxxz9EBnXdt8MYIy8vmyFFqCJRmy1UxLBYUN3TAyFftDysEenLh0mA1AYJKMCeEBw5IH8oPcIun
b6HXb4MtBWgEb11NwHcQiMCfysGigcVECKqBPwlxhw58KdFTxwTaPP6OiXUOW4cBCKiBw0gBQ9/SdxBb+oAgDNijybfcfrxLfdWROuTvlUCVYcgLC99vbcp0dPgq3cMH
8UzoHAjQ1UdhpcZWFqguZrb8KFFVZPeYKdxXTiYTknRmxtFoxOPjYwwGA3jv6b3fuhmpKkIICCFYlmWYTqfI87z6LxBaeqdp2titAt6GgFXGs+hAKYrC7u7uliXrVn+W
b8PV+A25Z0BSvbyERAAAAABJRU5ErkJggg==
}
image create photo ::img::update_27x24  -data {
iVBORw0KGgoAAAANSUhEUgAAABsAAAAYCAIAAACEIhGsAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH4wIYCQgGnQL7LgAABDtJREFUOMttVF1MXEUUPmfuvbAs
K7C7VCkluhYtPyuBlloIL6XUvkAwWkBTEkkrjySVhjcfxNrGF5QHrUaDpikPTWjoQ2NoYmKTAmmaBqsQSElVIuUn6ALl/87cn53jw12Wu1dOJjczZ+Z857tzvjMohAAA
InK+jLHEknN8/Fi9cweHh1nhSz2+z78eezMvLKuq6N137GPHZEYG2z9QCCGE4LsmhBBra8atW/FoVAJIAAL449zHtXVxAAJwfDIajQ8OGmtrqYFCcM6Bu00IPjNjtrXt
hQLQ8eJv3rvnD7t9iZ3WVnNujnuMwa4hYzg7q144r/X3E7hsdT1PXQ6H4H9GN2+qTc3q0hIyhntuXdcT/ObnrYaGfZgAGE0t7W8vuf/aPWprrVhMcJ7AYYwxRATLwuvX
1aEh8vBgTAaDaf/ONBdN1NRBMAiIXqr37yu9vZjAAQDnPsXkpNQ0b/ZAwG5sNPv7xfS02Fj7e9a4ccOsr7cDAe9BRZFTU8IwkpXZ2rIuXfKeCgatq1cF55yIW5YwhGVy
Im6aovtTMycn7jne3m7aNt9FXF2V+fkp++npdmcnl1I3TbcyOOemyYl4R4eVlpYSEQ7LjQ0HUdfF+LiHYLyiQjx9yqXk+5lt8/l5UVjopTkyIgyDM0RkU1MpV41ApSXy
yBEwDER07tvpDWduWVBQII8eJVV1lxCmppiiAAMiiMVSEP2ZdPhwsqRuLHfaaJQ0zSVOglgMAYA5x73apRS4JGJy6ZknfQDAiIgOHEjx7+ywP/9iAIC4L0FnOTHBOHc7
ITeXiIAhY1RW5tV1LKZIYMiS1NzQjIGx5UNTKX4VI/kIgAAgJZSXAxGgEAK4rr1RhouLiWyh3PGP3r/7VqSr8qLKFCtuuakxZCoqvQ++H/0thrY/XfUFZ9q/+zYjFKJn
z0xNA5WIQEuTra1qT48TuV5T8VXZzsDIZ3GDd1V3+dP9tm1LkgyZqqqGKT55cKVvvM+QW5qqIaNzZ1ZOPLpcViEzMsCyADnnQIQLC/hFjxXOWc5S7x7SLy72xbke8L1Q
+/LJlpKWmoKakD+0srPycOHhwPTA6PzotrGdLEZ2ZqDr0ODZ8lOvvS6JAHVdR8S4jP/8+0DHL52Z/px1EKvby8AQiBhTQr5QVnqWqqhW3No0Np/z50QyRR4mXWu89mF5
u+NTnddcYUpl6ZnKudqhyZ8gHcB57xAlyRV9eUVfTimqG86muqK65uIm54kkor0XNy/jxS/rek4V1YHh0TKmDLdZVP1KdV99X44vuKeTpDIIKJId+bHhh/NVF8Ai2EfA
qT1g0geVbQNnB/ID+ZQQGAE4lUntCm7ze7P3uoe7n/zzBBhAQnAIQOBkiUNJXvTKycunI6d9qs/TTonKJBGllM7EkMbY4tjt6dtjS7/Obs7qpp6Z5o9kR04crGoqbTp+
sFJDbd/A/wBQK+ZjS2v+MAAAAABJRU5ErkJggg==
}
image create photo ::img::us_32x32  -data {
iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAHDElEQVRYhcWXfXAU9RnHP7+9zeXucgdJLuRIwuUFmoCUYIWEl4BCm4CNCBVMywRqZyyjHVERplTaTrFO
qWUmo9NSpR2ndVpbbW1HO60vFWrVAj0FK5ggBkiwgQTSvF28l9ze7d7t/vrHHSGhHVFpJ8/M7rPP7s1+n9/v+d7zfVZIKZlIUzNemQBsa2wC1gQkAIDa1z+AJbKKk0bK
bloSKSU2RVGEIrAsC0tKFCFQFCUdWxJFGROPeS4tSY7LTn5uDpb1H6Udt0hFUUb6+vqGxMZNW7ce6lC+K6Swz6/2s3p5NXueOkB0JEFVRSEbVtXw2G8OMDQco7Qkn03r
FvH47wP0DkTwed1sbr6eJ//0Fl3nh7FSKdbdeB0PfX0tWtz40JULiGzfvn2jeuzUQHPW5Fm5c2aXgU3hzLCBr6KMUtWGqip0DuoU+P0UlStI4MygTl5xMd6SElKWpHNQ
J6fQx5wCHwOhEXylfux2J3a780q777K7nI2CSTcEGjduqLuzaTFf2fUc0ZDGstoZfOPLS7hz9/P0DoSpmT2NXV+r595HXuJMT5DZFVP44X2N3L93P22d/ZQV5fLT+1fz
7ScOMCs8TMt0FzriEpTMnCQg094yEjx44I8Pi6LqdQH7tOvqfIW5hCIa00vyeL9nGJ/XzVAoRqXfy+lzQ0wt8NAfjFJVVkDHuSGmej30BaNUlRbQ2RPEl++mI6ixvPuf
fPPMu2iKbUwCMpODBMtCSEFSC/OwdbZFWb6wkrCW4K3j3VQUTWbzugVocZ3A8W58eTlsvnUBUkoCbd1Mctm559aFOOw2Am3d2BS4u2kh+R4Hb77bQySi4SgqwNVQi7u+
BvfnMsfodS3u+gW462vxrKwju6QYYfffGLh26Yq6hkVVPPnSO5imRXlRLo2Lq3hqXytxPYkv38Mty67hmf3HiWg6uR4H6xuqefb19wiGNHIcdjZ8/lp+u7+NlYur2LNj
3ZXqD8C9W7a0KEbcIMum4J3sIpVMEUsY2BRBQa6LVMoknkgikBTkujAtCy1hIC0Lb24OwpJoCQPLMpmS50K1CaTy0XuaUG2INRu3BQI9zrpgOEHDghnc/oUaduz5M/3D
MRbPmcY9zUv4zmP7OdcXovpTPnbc/lm+/7PXOH1ukOnFeTx410oe+eXrtJ7uxebMpilXsNMn0aUAkSaiEAIQCNL3hBRYSZ1dh19uUc9eCGK3lVLpz6dvKMy+wClURTCz
1MtwSGPf308hLcnMsgJimsHLh9qJjWhU+DzEIiM8/8o7DA2GKCt0869YimhHF8G/tKIpKqM6c9FbEhAoQoBpEqcXW39o8qbrb1jo33lHPb968RhHWs9RO8fP9zav4Nm/
HudvR7uYVe5l95ZGXni1jVcPtFOY7+bH31rLwX+c4ZWD7Xg8TvY+8CWOnbqAV5qsmj+drJkVuGZW4Kwqxznqy3FUleGsLMVZVc5hIxwQFbXrA4ansm5SjgOJZM6MqRxt
P88kdzaGYTK3ciqHW7twu+xocYN5n/Zz9EQPblc2kViC2upSWtvP43TaOXs+yG1rF/GTB5o/Ege2bL2vRamdW85ILMHJjl6KvB6aVsxFi+u0nbyAx6HStKIaM2XxXkcf
Toed5ptqyMqycaKjF0URNN9Ug8ft4ERnH7F4EkV8HGEVCFvxysD8pfV1q5bN5tGnD5EwUswsLWBtQzV7nz7EiKbjL8rjtjW1PP67NwiGYhR63dzxxcX84g9H6B2IkDvJ
yV3NS3nixaPMiwTZVSjRMxCjkiTluENI+MHxgy2qGTcIR+MMh2OEo3GSCYNQxEkoohGN6YzEDaIxnVBUIxqLE9V0nA6VDyJxorEEUU3HZhOEIhojeorw+z2E3zhJLNMJ
xxHRtJASBCYCSYJBRO3KTYHi6fPqkikTV7aKEGmy6noSl9OOEALLkiT0JE6HHSHSL40nkjgcWWlGS9ASBsKRzcJklCatH310xpGXtMBKt2QhJUJKHnr7tRZ15903s3rN
LR+jbv87m7JtG4phTtgwhDRN1K7dj9L26xcyXSvTvTIdLK2oIu0vsumiyv63+PJnYwM5Vo4lAhhqexM12dFOovc8UggQSvotyhjgsbr+ydY5bg4ACywLG2CGB1DzGpbg
m30NkjErHwW/2gTkZZdyNBFFkeTs249asumrlDfefBUgn9zytDiq1OOXxGJ88f6PltnVpIEafuZH6sCJ5zBNLivDpd+NLYL8kPhKPBzbE7KzQD96TFXf7jzZvXRKcIGC
kk5gFPFqyXe5XQJXgJFUilPdFy4Ij8f9mVK3XG8T0qUoCkJR0qODEJnNuLpEZIZ4Usr0TCotsExFl1mDXR8kfy4OHz6CnrIwkknFZrOlgQFEZnC4Skv/A+WoJliWhWma
uHNclpHQEBP9dTwRX8Xj7N+2ai+CDrtU9gAAAABJRU5ErkJggg==
}

image create photo ::img::delete -format GIF -data {
    R0lGODlhEAAQAIABAIQAAP///yH5BAEAAAEALAAAAAAQABAAAAIjjI+pmwAc3HGy
    PUSvqYpuvWQg40FfSVacBa5nN6JYDI3mzRQAOw==
}

image create photo ::img::print -format GIF -data {
    R0lGODlhEAAQALMKAAAAAP///52VunNkl8C82Yl+qldBgq+pyrOzs1fYAP///wAA
    AAAAAAAAAAAAAAAAACH5BAEAAAoALAAAAAAQABAAAARGUMlJKwU4AztB+ODGeUiJ
    fGLlgeEYmGWQXmx7aXgmAUTv/74N4EAsGhOJg1DAbDqbwoJ0Sp0KB9isNis0eL/g
    ryhH5pgnEQA7
}

image create photo ::img::view -format GIF -data {
    R0lGODlhEAAQAKIHAP///wwMDAAAAMDAwNnZ2SYmJmZmZv///yH5BAEAAAcALAAA
    AAAQABAAAANMKLos90+ASamDRxJCgw9YVnlDOXiQBgRDBRgHKE6sW8QR3doPKK27
    yg33q/GIOhdg6OsEJzeZykiBSUcs06e56Xx6np8ScIkFGuhQAgA7
}

image create photo ::img::avec_det -data {
iVBORw0KGgoAAAANSUhEUgAAAAgAAAAICAIAAABLbSncAAAACXBIWXMAAA7EAAAO
xAGVKw4bAAAAB3RJTUUH4wEVCRAgeN5EhwAAAB1pVFh0Q29tbWVudAAAAAAAQ3Jl
YXRlZCB3aXRoIEdJTVBkLmUHAAAAJklEQVQI12P8//8/AzbAxEAywGrU////mTDl
IFwmNH1wBiPJrgIABS4X85g6J8IAAAAASUVORK5CYII=
}

image create photo ::img::avec_det1 -data {
iVBORw0KGgoAAAANSUhEUgAAAAoAAAAFCAYAAAB8ZH1oAAAABmJLR0QA/wD/AP+g
vaeTAAAACXBIWXMAAA7EAAAOxAGVKw4bAAAAB3RJTUUH4wEUEhEE9Qm64wAAAB1p
VFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoIEdJTVBkLmUHAAAAQElEQVQI133K
MQrAMAwEwTXkzyqlUj92uakCSiA+2OqG7hY4VlWicsJVpQr6jx/0gl88kcrFWESs
vbcAmbnmdwPsWlLtRYEjcgAAAABJRU5ErkJggg==
}

image create photo ::img::not_det -data {
iVBORw0KGgoAAAANSUhEUgAAAAgAAAAICAIAAABLbSncAAAACXBIWXMAAA7EAAAO
xAGVKw4bAAAAB3RJTUUH4wEVCQ4Al/FbkAAAAB1pVFh0Q29tbWVudAAAAAAAQ3Jl
YXRlZCB3aXRoIEdJTVBkLmUHAAAAMUlEQVQI12P8//8/AwMDIyMjhAEHTHAWIyMj
dgk0ORQJZDl0CZw64E5gwiqKIoHmXADVjg8Nbya7dgAAAABJRU5ErkJggg==
}

image create photo ::img::not_det1  -data {
iVBORw0KGgoAAAANSUhEUgAAAAUAAAAKCAYAAAB8OZQwAAAABmJLR0QA/wD/AP+g
vaeTAAAACXBIWXMAAA7EAAAOxAGVKw4bAAAAB3RJTUUH4wEUEhIjey5cSwAAAB1p
VFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoIEdJTVBkLmUHAAAAQUlEQVQI12WO
MQ7AQAjD3G68OCP8PF1axNFILLaIgiTbZh6AtwDwFg2nuF7QkcTNSkSc71Xlo/MD
DSewDZn5G/8A70VdxL6tU7QAAAAASUVORK5CYII=
}

image create photo ::img::logo  -data {
iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH4wIaDhUdR02A0AAAAB1p
VFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoIEdJTVBkLmUHAAAIMklEQVRYw8WXfXBU1RnGf+fcu5u7+f5OSAiQkBCooJQqRfEDwWhHEGupreLUtjPaD7XTTq3TsVZt
aWfaYSydqdqpykCrVesInQ5MUWmrFhBbFARiMCpGQCAhIYTsZje7e+85b/+4m4QgOlO10zNzZ/beM3vO+z7v8z7nOUpEhP/jcAGeerGTtZu7cF0QQI2EpP43m/qBcOvS
VpbOaw4DyIpLRjyM1bkI1GmRfLShCJeB8IcCRCmMWHyJjCEQUZDnalytAIU1FoNCCN8V8jGCUCgxKA2OdgAhQOPk4HUBrAUTWLSj8a3lqwvrWTS7Djs+h488NLBxx3us
33YMR4MxAjnquQAiFmvDrK0V8hyH/IIoYj8ZfmqtcF0HayyIwlo7Si8XAAtiBVEWV8Ndv9/JzateRJT+2JsrwrXrq4qYWFOBFYMKFGYcAghiDTYHd0t9KTMmlY1yUD6B
IHyjyAQ+oDCixpcAO/qObxS3XD2Fay5sxFgQEXSuKSRXz7BLcn/QzhjNrQ3nRkJWzigRn9i8n9XPvoPrhPNaTimBRSFheVAiHDqa4M11m5my/jH68vLZe+mVtD3zNEnH
Zd0Xv8bUhia8WAxRislbn6XijT10TmziwJz5VFXVoHRYurqd26jdtZ2e8gp2lczCmiiiFGLt+BJoaxHfgAgumgf+sp+B7D5+VBLjbwd6ePH+1Vw+u47Xek7wjdtWUHzW
MhyvlAlBnPaaV/GdCBv/vIGfrd5GwdTL0E6EfJPljcrtKNdhT2cXjxx6jTmXLAcKkMCirJyKgBAAjhUChFnNNVza244ayrIvPsi8sgpMJsvuRILWyVWcO6+ZapNlxaHn
yVjQEuCpLDNqI8w+r4EShHu61hHRCqzFVYaSiI8xgrYWK3Y8BxTgGgGlWBrfy9fppijZh6RS3FRbRqUXgVSKZaX5VLc2cfWXJuLddTdBnjeqbkVRzbIFU7n37kX4y64l
8GIQhBLmKcuU2gKMMURERrMf6wIrYV0M1CZ7qexpJxXSjemOw3A8iwEmRyL8+oWX2NvtszKb4ZWTCapdh4leHkPxJC/sfA95ZDs/Sac4kEqTMpZp+R7pZIojSYe6esFa
wdgxIdIjbWhyPH+6fC5t5mycwQGy8Titr75FYjBOQSrFOTve4Kmjaf7Z5XMnM5nfFZAZDpBkkvTQMHvfHuDJZw+wMjqHaZ0Z+pPhXDaRYiCRDbtDwkfGCVHu3FEiDDge
Z8+Zi7NhIz1OlGRtNdXJIbqB/rIS2q78Cg3NswG4beHFFP71txDvI4Jh5tQyLrm8kQSNfGf+PKq3PIH0HsD1M5QVRmE089NKELaxgCg+272bm4J20slhCiIBa8qipPvj
OFrzQFkpMxdMofmqC0GHp1n6+d9hD8cpiEZYNHcS937rAlDhstk3N+DvjxNRivr6GIEVIqPoy/gARMJP046/w8TOv2O0Q3k6zQVDQ2SVIh9YkhzilttX0LF2BxMaz8WP
5PGLd4/RkExyIuHw5KYOdh9fhxvJI9AOd+w+yMxkkiED+w/HmVITCtv7EFCSi8Ba/jH5AhrffZlpg4f5kxOjTgkX2wzPqAjPuUXsiftkuuOkvCwmz6E/kaIwmSSpPU4m
LQd6AxxXE2jh+OAwJ5JJEqLxDTgj+4xkO64EuXY8WDIZY9P0+z7b6yaxPJXkWF+cXdVleJddyQ8XLaayZgIbtvXzVl/AQDZLNJMhFXGJug6FsQjaiZKymkSQoSeVYtCJ
ovWYvQr5NnZUhxqeg0ZZn+zAUfqMId3YQHRokD5rOV5VwZIr5vPNG5ew7IpzqSj2EAMnjc8xYxgWIRZVVBRH+Mz0Eh655yIGMml6fJ+4sePdnYBVZ+KAhfwgyXPV8/AH
utj6ei/1biNu7ATPH0nT+XA7D278AwCOV0DEzWPrWVeRfG8XHfFeFsxr4sYbz6HrSIotWzvxvrycrv072b6rHUd5YaOL5JTwVCHKmRKxiqwTZev5N3P8cAelQZYt5U2k
Ez3EskP4XgMnhx0EKHXBJWDf1IXEWmbxy+tb2fzSQQJjWb74LLSjiF/zaVY9tJEln7+Bq6Mej26OI8aGwne6FIeIaAJj+P4N59E2fxkyYiQllGmlwqPVdTV3P7iNjrcH
0Y6DUI3v1nHh+VVcMrcll5KipDDKT2+/lh+v2kTzlFqEFKDQolG56o8q4agBFog6QnFMUxTTFMYcivIdimIOhZ6mwFMUeBpHKUQgmw74+ffmgsCnmmvO6Ocb6iq4bvEs
ZreWgGgEGa8DIqEDFsB1HFY+vIN7f/UCSo044vHm1IpQOqESz/OIRhTfvnMTq+5pY+frR5gxtXoUAdBYsfQeH2LN06/w7919FBbnh3Pj2jCnAWLDj0UVpZRUFCF8sCe0
Eto4AkVTy0QmTyimY98h1m/aw5K2mURdTXdfgocfe5nFbTPoOhrHy/cQa0O+nS7FIwLlG/ju9S0sXdjCB13aIhGHH6zcws7OflwF3ceSXHTtH7nvrgW0NlWz5ql/4fuW
okKPO25dwBduepyDvZbKCWXYEef2vrMAQYmgEXp6E7R3dJ/RjIZl0qRSmZA3SuG4mqbpDTzw6C5SmTTTJlVy3ZLprPjNDh56/BVUfgVVtQ5iQ0w1p3XByMqS8/Br1u/j
/rWv5gzmGYIQS2VNJV5+BJtbSKxgY4UUeKUc7M9y56qXqa6vJPANjtacesVQIuMRsFbwT1Gr4pJSSkvLP7D+AlgJCPwcddWYxFoCtNKUV5fjZwwKMCPCk+NyYBgNyAWY
0VTKdZ9rQGv935n9D7s0fMi8tUJLQxEA/wFxuQEKdcuI4gAAAABJRU5ErkJggg==
}
#-----------------------------------------------------------------------------
# asnGetUTCTime: Extract an UTC Time string from the data. Returns a string
#                representing an UTC Time.
#
#-----------------------------------------------------------------------------

proc edithex {hex} {
    set c ""
    set l [string length $hex]
#    puts $l
    for {set j 0 } { $j < $l} {incr j +2} {
	set c "$c[string range $hex $j $j+1] "
    }
    return [string toupper [string trimright $c]]
}

proc aboutUtil {w type} {
    global nickTok
    if {$type == 0} {
	set title [mc "Parse and view the X509 certificate from token PKCS#11"]
    } else {
	set title [mc "Token mechanisms"]
    }
    set content "
	Copyright (c) 2019 Vladimir Orlov 
	email: <vorlov@lissi.ru>
	http://soft.lissi.ru/
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See the GNU General Public License for more details.
	
	[mc {The distribution is here}] (Linux, OS/X, Windows):
	
	http://soft.lissi.ru/ls_product/utils/chainfromcert/
	"
    set content_ru {

    }
    
    catch {destroy $w}
    toplevel $w -bg #93cee9
#    toplevel $w -padx 4 -pady 4 -bg #c0bab4 
    wm geometry $w +150+100

    wm title $w $title
    wm iconphoto $w ::img::logo
#    $w configure -background #39b5da
#    $w configure -background #c0bab4
    
    text $w.text -width 66 -height 20 -background #ffffff -yscrollcommand [list $w.vsb set] -font {Times 10 bold italic}

    ttk::scrollbar $w.vsb -orient vertical -command [list $w.text yview]

    $w.text tag configure tagAbout -foreground blue -font {Times 10 bold italic}
# "helvetica 13 bold roman"
    $w.text image create end -image creator_small

#    lightgray
    frame $w.separ -height 5 -bd 2 -relief groove -bg #e0dfde
#-class WizSeparator 
    frame $w.butt -bg #f5f5f5 -highlightthickness 3 -highlightbackground #93cee9 -highlightcolor #93cee9
#    frame $w.butt  -bd 3 -relief groove -bg #c0bab4
    #ffffd1
#    $w.butt configure -background #39b5da
    pack $w.butt -expand 1 -fill both -side bottom
    pack $w.separ -expand 1 -fill x -side bottom -pady 0
    pack $w.vsb -side right -fill y -in $w
    pack $w.text -padx 3 -pady {0 0} -side left
    ttk::button $w.butt.ok -text "Ok" -command "destroy $w"
    pack $w.butt.ok -side right -padx {0 10} -pady 4 -in $w.butt
    if {$type == 0} {
	$w.text insert end "\t$title" tagAbout
	$w.text insert end $content
    } else {
	set llmech [pki::pkcs11::listmechs $::handle $::slotid_tek]
	$w.text insert end "\t$title: " tagAbout
	$w.text insert end "$nickTok\n"
	foreach mech $llmech {
	    $w.text insert end $mech
	    $w.text insert end "\n"
	}
    }
    $w.text configure -state disabled
    
    # do not allow closing via [x] button
    wm protocol $w WM_DELETE_WINDOW ";"
    
    # now get input
    grab set $w
    tkwait window $w
#    return $input_values(exit)
}
#Список токенов со слотами
proc listts {handle} {
    set slots [pki::pkcs11::listslots $handle]
    puts "Slots: $slots"

    set listtok {}
    foreach slotinfo $slots {
	set slotid [lindex $slotinfo 0]
	set slotlabel [lindex $slotinfo 1]
	set slotflags [lindex $slotinfo 2]

	if {[lsearch -exact $slotflags TOKEN_PRESENT] != -1} {
	    lappend listtok $slotlabel
	    lappend listtok $slotid
	}
    }
#Список найденных токенов в слотах
    puts $listtok
    return $listtok
}
proc selectTok {w nametok} {
    puts "selectTok=\"$nametok\""
    set txt ".text.text"
    set lists [listts $::handle]
    if {[llength $lists] == 0} {
	tk_messageBox -title [mc "PKCS#11 library"] -icon info -message "[mc {Tokens not present}]\n" 
	$txt delete 0.0 end
	.saveCert.labExp.listCert configure -state normal
	.saveCert.labExp.listCert delete 0 end
	.saveCert.labExp.listCert configure -state readonly
	.selecttok.tok configure -state normal
	.selecttok.tok delete 0 end
	.selecttok.tok configure -state readonly

	return
    }
    foreach {lab slotid} $lists {
	if {$lab == $nametok } {
	    break
	} 
    }
    .text.text delete 0.0 end
    .saveCert.labExp.listCert configure -state normal
    .saveCert.labExp.listCert delete 0 end
    set ::listx509 {}
    set ::listx509 [listcerttok $::handle $lab $slotid]
    set ::slotid_tek  $slotid
    .saveCert.labExp.listCert configure -values $::listx509
    .saveCert.labExp.listCert delete 0 end
    .saveCert.labExp.listCert insert end [lindex $::listx509 0]
    .saveCert.labExp.listCert configure -state readonly

    set nick [lindex $::listx509 0]
    puts "selectTok NICK=$nick"
    if {[llength $::listx509]} {
	$txt delete 0.0 end
	::fileWithCert $txt $nick $::arrayCer($nick)
    }
}

proc cert_valid_date {} {
    # check validity date
    set startdate $::notbefore
    set enddate $::notafter
    # today
    set now [clock seconds]
    
    # <now> should be between start/end dates
    set isvalid 1
    set reason "Certificate is valid"
    #puts "validity: startdate=$startdate, enddate=$enddate, now=$now"
    if {$startdate > $now} {
        set isvalid 0
        set reason "Certificate is not yet valid"
    }
    if {$now > $enddate} {
        set isvalid 0
        set reason "Certificate has expired"
    }
    set result [list $isvalid $reason]
    return $result
}


proc validaty_cert_from_crl {crl sernum issuer} {
#puts "SERN=$sernum"
    array set ret [list]
    if { [string range $crl 0 9 ] == "-----BEGIN" } {
	array set parsed_crl [::pki::_parse_pem $crl "-----BEGIN X509 CRL-----" "-----END X509 CRL-----"]
	set crl $parsed_crl(data)
    }
    ::asn::asnGetSequence crl crl_seq
	::asn::asnGetSequence crl_seq crl_base
	    ::asn::asnPeekByte crl_base peek_tag
	if {$peek_tag == 0x02} {
		# Version number is optional, if missing assumed to be value of 0
		::asn::asnGetInteger crl_base ret(version)
		incr ret(version)
	} else {
		set ret(version) 1
	}

	::asn::asnGetSequence crl_base crl_full
		::asn::asnGetObjectIdentifier crl_full ret(signtype) 
	    ::::asn::asnGetSequence crl_base crl_issue
		set ret(issue) [::pki::x509::_dn_to_string $crl_issue]
		if {$ret(issue) != $issuer } {
		    set ret(error) "Bad Issuer"
		    return [array get ret]
		}
		binary scan  $crl_issue H*  ret(issue_hex)
	    ::asn::asnGetUTCTime crl_base ret(publishDate)
	    ::asn::asnGetUTCTime crl_base ret(nextDate)
#puts "publishDate=$ret(publishDate)"    
#Список сертификатов отозванных
	::asn::asnPeekByte crl_base peek_tag
	if {$peek_tag != 0x30} {
#Список сертификатов отозванных пустой
	    return [array get ret]
	}

	::asn::asnGetSequence crl_base lcert
#	binary scan  $lcert H*  ret(lcert)
	while {$lcert != ""} {
	    ::asn::asnGetSequence lcert lcerti
#Разбираем i-й сертификат
		::asn::asnGetBigInteger lcerti ret(sernumrev)
		set ret(sernumrev) [::math::bignum::tostr $ret(sernumrev)]
#puts "$sernum $ret(sernumrev)"
		if {$ret(sernumrev) != $sernum} {
		    continue
		}
		::asn::asnGetUTCTime lcerti ret(revokeDate)
		if {$lcerti == ""} {
		    puts "NO reasone"
		} else {
		    ::asn::asnGetSequence lcerti lcertir
		    ::asn::asnGetSequence lcertir reasone 
			::asn::asnGetObjectIdentifier reasone ret(reasone) 
			::asn::asnGetOctetString reasone reasone2
			::asn::asnGetEnumeration reasone2 ret(reasoneData)
		}
	    break;	
	}
    return [array get ret]
}

proc selectCert {w nick} {
    set txt ".text.text"
    puts "selectCert NICK=$nick"
    $txt delete 0.0 end
    ::fileWithCert $txt $nick $::arrayCer($nick)
}


#Список сертификатов
proc listcerttok {handle token_slotlabel token_slotid} {
#puts "token_slotlabel=$token_slotlabel"
#puts "token_slotid=$token_slotid"

	set ::certs_p11 [pki::pkcs11::listcerts $handle $token_slotid]
	if {[llength $::certs_p11] == 0} {
	   tk_messageBox -title "[mc {Export/View certificates}]" \
	 -icon info -message "[mc {Certificates are not on the token}]: " -detail "$token_slotlabel"
	    return
	}
#puts $certs

	set listCer {}
	array set ::arrayCer []
	foreach certinfo_list $::certs_p11 {
	    unset -nocomplain certinfo
	    array set certinfo $certinfo_list
	    set certinfo(pubkeyinfo) [::pki::x509::parse_cert_pubkeyinfo $certinfo(cert)]
#parray certinfo
#	puts "Cert: $certinfo(pkcs11_label) / $certinfo(subject)"
	    set ::arrayCer($certinfo(pkcs11_label)) $certinfo(cert)
    	    lappend listCer $certinfo(pkcs11_label)
	}
	return $listCer
}

proc readPw ent {
	global widget
	global yespas
	global pass
	puts "readPWD"
	set pass [$ent get]
	puts $pass
	$ent delete 0 end
	set yespas "yes"
	wm state .topPinPw withdraw
	
}

proc saveCertFile { w list cert } {
    #package require base64
    global home
    global yespas
    global pass
    global tekSlot
    global libpkcs11
    global p11conf
    global listCer
    global formatCert
    global nickTok
puts "tekcert=$::tekcert"
	puts "$cert\n"
	if {$::dercert == "" } {
	    tk_messageBox -title [mc "Export certificate"] -icon info -message [mc "No certificate selected"]
	    return
	}
	if {$::tekcert == "file"} {
	    puts "nickTok=$nickTok"
	    set lists [listts $::handle]
	    if {[llength $lists] == 0} {
		tk_messageBox -title [mc "PKCS#11 library"] -icon info -message "[mc {Tokens not present}]\n" 
		return
	    }
#DER в бинарнике
	    binary scan $::dercert H* cert_der_hex
#Ввод PIN-кода
	    set yespas ""
	    set pass ""
	    set titpin "[mc {Token}]: $nickTok"
	    wm title .topPinPw $titpin
	    wm state .topPinPw normal
	    wm state .topPinPw withdraw
	    wm state .topPinPw normal
	    raise .topPinPw
	    grab .topPinPw
	    focus .topPinPw.labFrPw.entryPw
	    set yespas ""
	    vwait yespas
	    grab release .topPinPw
	    if { $yespas == "no" } {
		return
	    }
	    set yesno "no"

	    set labcert "$::cn_subject from $::cn_issuer"
	    set uu [dict create pkcs11_handle $::handle]
	    dict set uu pkcs11_slotid $::slotid_tek
	    lappend uu "pkcs11_label"
	    lappend uu $labcert
	    puts "LISTforP11=$uu"
	    set pkcs11id [pki::pkcs11::importcert $cert_der_hex $uu]
	    unset uu
	    set uu [dict create pkcs11_handle $::handle]
	    dict set uu pkcs11_slotid $::slotid_tek
	    lappend uu "pkcs11_id"
	    lappend uu $pkcs11id
	    lappend uu "pkcs11_label"
	    lappend uu "$labcert"
	    if { [pki::pkcs11::login $::handle $::slotid_tek $pass] == 0 } {
		tk_messageBox -title [mc "Import certificate"] -icon info -message "[mc {Bad PIN}]"
		return
	    }
	    pki::pkcs11::rename all $uu
pki::pkcs11::logout $::handle $::slotid_tek
puts "Установлена метка $labcert"
	    
	    updatetok
	    tk_messageBox -title [mc "Import certificate"] -icon info -message "[mc {Import certificate}] \n$pkcs11id" -detail "$labcert"
	    return
	}
	set typeCertDer {
	    {"[mc {Certificate to DER format}]"    .der}
	    {"[mc {Certificate to DER format}]"    .cer}
	    {"[mc {All}]"    *}
	}
	set typeCertPem {
	    {"[mc {Certificate to PEM format}]"    .pem}
	    {"[mc {Certificate to PEM format}]"    .crt}
	    {"[mc {All}]"    *}
	}
	if { $formatCert == 1 } {
	   set typeCert [subst $typeCertPem]
	   set typeTitle [mc "Export certificate to PEM format"]
	} else {
	   set typeCert [subst $typeCertDer]
	   set typeTitle [mc "Export certificate to DER format"]
	}

	set file [tk_getSaveFile -title "$typeTitle"  -filetypes $typeCert -parent .saveCert -initialdir "$home"]
	if { $file == "" } {
	    return;
	}
	set x [catch {set fid [open $file w+]}]
	set y [catch {puts $fid "#http://soft.lissi.ru\n"}]
	set z [catch {close $fid}]
	if { $x || $y || $z || ![file exists $file] || ![file isfile $file] || ![file readable $file] } {
	    tk_messageBox -parent $w -icon error  -message "Невозможно записать сертификат в этот файл:\n \"$file\""
	    return
	} 
	file delete -force $file
	#Собственно запись в файл
#Разобраться с PEM
	set encodePem $formatCert
	if {$encodePem} {
		set cert [::pki::_encode_pem $::dercert "-----BEGIN CERTIFICATE-----" "-----END CERTIFICATE-----"]
	} else {
	    set cert $::dercert
	}
	set fd [open $file w]
	chan configure $fd -translation binary
	puts -nonewline $fd $cert
	close $fd

	if {![file exists $file]} {
	   tk_messageBox -title [mc "Export certificate"] -icon error -message "Экспорт сертификата не удался в файл\n $file"
	    return
	}
	if {![file size $file]} {
	   tk_messageBox -title [mc "Export certificate"] -icon error -message "Экспорт сертификата не удался\n $file"
	    return
	}
	tk_messageBox -title [mc "Export certificate"] -parent .saveCert -icon info  -message "Сертификат сохранен в файле\n\"$file\""
}

#ttk::style theme use clam
#ttk::style theme use clearlooks
ttk::style configure TLabel -background #ffffff
#TkDefaultFont   -family sans-serif -size 12 -weight normal -slant roman -underline 0 -overstrike 0
set myfont [font configure TkDefaultFont]
set myfont [lreplace $myfont 5 5 bold]
set com "font create TkDefaultFontBold $myfont"
set com [subst $com]
eval $com
set myfont [lreplace $myfont 3 3 12]
set com "font create TkFontForFirst $myfont"
set com [subst $com]
eval $com
#set a1 [font configure TkFontForFirst]
#puts $a1

set w ".text"
#catch {destroy $w}
#toplevel $w
wm geometry . 78x20+150+100
wm title . [mc "Parse and view the X509 certificate from token PKCS#11"]
#wm iconname $w "text"
wm minsize . 47 16
wm iconphoto . ::img::logo
#frame $w -bd 1 -bg #b6b6b3
frame $w -bd 1 -bg #ffffff
pack $w -expand 1 -fill both 


#wm minsize .text 492 443
## Choose PKCS#11 library
labelframe .fr3 -relief flat -text [mc "Choose a PKCS#11 library for you token"] -font TkDefaultFontBold -bg snow -bd 3 -pady 3 -highlightthickness 2 -highlightbackground #93cee9  -highlightcolor #93cee9
label .fr3.lablib -relief flat -text "[mc {PKCS#11 library}]:" -font TkDefaultFontBold -bg #ffffff -anchor w -padx 0
# -image img_p11 -compound left
pack .fr3.lablib -in .fr3 -side left -expand 0 -fill none -padx 5 -ipady 4 -pady {0 4}
entry .fr3.ent -background snow -state readonly -width 40 -highlightthickness 1 -highlightbackground #93cee9  -highlightcolor #93cee9 -relief flat 
pack .fr3.ent -in .fr3 -side left -expand 1 -fill x -padx 5 -ipady 1 -pady {0 4}

#ttk::style map MeB.TButton -background [list !pressed snow pressed snow  active #93cee9]
ttk::style map MeB.TMenubutton -background [list disabled #ffffff pressed #ffffff  active #93cee9] 
ttk::style configure MyBorder.TButton -background snow -foregrpund snow  -borderwidth 0
#ttk::button .fr3.browser -image openfile  -compound center -command {fileDialog . .fr3.ent "so" .fr1.fr2_list} -state normal -padding 0 -width 1 -style MeB.TButton
button .fr3.browser -image openfile  -compound center -command {fileDialog . .fr3.ent "so" .fr1.fr2_list} -state normal  -compound left -bd 0 -background #ffffff -activebackground #ffffff -highlightthickness 0

# -style MyBorder.TButton
# -text " [mc {Select library}]"
# -width 20 
label .bh -text [mc "Choose a PKCS#11 library for you token"]
.bh config -bg yellow
bind .fr3.browser <Enter> {.bh configure -text [mc "Choose a PKCS#11 library for you Token"];place .bh -in .fr3.lablib -relx 1.0 -rely 1.0}
#bind .fr3.browser <Enter> {.bh configure -text [mc "Choose a PKCS#11 library for you Token"];place .bh -in .fr3.ent -relx 0.50 -rely 1.0}
bind .fr3.browser <Leave> {place forget .bh}


pack .fr3.browser -in .fr3 -side left -padx 5 -pady {0 4}
pack .fr3 -side bottom -expand 1 -fill x




global libpkcs11 
set libpkcs11 ""
global avecdet
#set avecdet 0
set avecdet 1

proc fileDialog { w ent  type list } {
    global typesys
    global p11conf
    global libpkcs11
    global tclpkcs11
    global home
	
	set lib11 [mc {PKCS#11 library}]
	set allf [mc {All}]
	set typesOth { 
		{"$lib11"		{.dylib}	} 
		{"$allf"		*} 
	}
	set typesX11 { 
		{"$lib11"		{.so .so.*}	} 
		{"$allf"		*} 
	}
	set typesWin32 { 
		{"$lib11"		{.dll}} 
		{"$allf"		*} 
	}
	switch [tk windowingsystem] {
		classic - aqua  {
			set  types $typesOth
		}
		x11 {
			set  types $typesX11
		}
		win32 {
			set  types $typesWin32
		}
		default {
			set  types $typesOth
		}
	}	
	set types [subst $types]
	puts "$type"
	
	if { $type != "so"} {
	    return ""
	} 
	switch -- $::tcl_platform(platform) {
	    "windows1"        { 
		set file [encoding convertfrom cp1251 [tk_getOpenFile -title [mc "PKCS#11 library"]  -filetypes $types -parent $w -initialdir $home]]
	    }
	    "unix" - default { 
		set file [tk_getOpenFile -title [mc "PKCS#11 library"]  -filetypes $types -parent $w -initialdir $home]
	    }
	}
	puts "FILE=$file"
	puts "LIBFILE=$libpkcs11"
	
	if {[string compare $file ""]} {
#		$ent delete 0 end
#		$ent insert 0 $file
#		$ent xview end
#		puts "Загружаем библиотеку, ищем заполненные слоты"
		.fr3.ent configure -state normal
		.fr3.ent delete 0 end
		.fr3.ent insert end $file
		.fr3.ent configure -state readonly
		set ::pkcs11_module $file
#		set ::pkcs11_module "[encoding convertto cp1251 $file]"
		updatetok
	} else {
	    set file [$ent get]
	    puts "GET=$file"
	}
}
proc InfoMech { w list } {
    encoding system utf-8
    global libpkcs11
    global labTok
    global tekSlot
    global p11conf
    pack forget .saveCert
#    place forget .saveCert
	    $list  delete 0.0 end;
	    if { $libpkcs11 == "" } {
		set libpkcs11 [fileDialog . .fr3.ent];puts "res1=\"$libpkcs11\""
	    }
	    if { $libpkcs11 == "" } {
		set resTok "Не быбрана библиотека PKCS11!"	
		$list insert 0.0 $resTok
		return 0
	    }
	InfoSlot $w $list $libpkcs11
	$list delete 0.0 end
	if {$labTok == ""} {
#		 tk_messageBox -title "Slot and token" -icon info -message "[mc {Found no slots with tokens}]"
	        return
	}

	puts "LIBP11=\"$libpkcs11\""
	    set command "\"$p11conf\"  -A \"$libpkcs11\"    -m -c $tekSlot"
	    set tube [ open |$command r+] 
	    while {![eof $tube]} {
		gets $tube res
		$list insert end "$res\n"
	    }
	if {[catch {close $tube} result]} {
	   tk_messageBox -title "Механизмы токена"  -icon error \
	   -message "Ошибка доступа к  токену\n$result"         
	        return
	 }
}


proc changelang {w} {
    global loadfile
    global lang
    if {$lang == "ru"} {
	set lang "en"
	.saveCert.labExp.but.lang configure -image  ::img::us_32x32 -compound left
    } else {
	set lang "ru"
	.saveCert.labExp.but.lang configure -image  ::img::ru_32x32 -compound left
    }
    focus -force $w.text
    ::msgcat::mclocale $lang
    set title [mc "Parse and view the X509 certificate from token PKCS#11"]
    wm title . $title
    .selecttok.lab configure -text "[mc {Select token/smartcard}]:"
    .fr3 configure -text [mc "Choose a PKCS#11 library for you token"]
    .fr3.lablib configure -text "[mc {PKCS#11 library}]:"

    .topPinPw.labFrPw configure -text "[mc {Enter the PIN code and press Enter}]"
    .topPinPw.labFrPw.butPw  configure -text [mc "Cancel"]

    .saveCert.labExp.but.dopcom.menu entryconfigure 0 -label [mc "Load chain CA"]
    .saveCert.labExp.but.dopcom.menu entryconfigure 1 -label [mc "Validaty by CRL"]
    .saveCert.labExp.but.dopcom.menu entryconfigure 2 -label [mc "Save certificate"]
    .saveCert.labExp.but.dopcom.menu entryconfigure 3 -label [mc "OCSP validity"]
    .saveCert.labExp.but.dopcom.menu entryconfigure 4 -label [mc "Verify signature"]
    .saveCert.labExp.but.dopcom.menu entryconfigure 6 -label [mc "About"]

    .saveCert.labExp.but.dopcom configure -text [mc "Additionaly"]
    .saveCert.labExp.but.butCan configure -text [mc "Exit"]
    .saveCert.labExp.but.loadcert configure -text [mc "Load certificate"]
    $w.tit.left.caption configure -text [mc "Details"]
    .saveCert.labExp.labCert configure -text "[mc {Select certificate to view, import, export or remove}]:"
    .saveCert.labExp.typeCert.lab configure -text "[mc {Choose format to export}]"
    .saveCert.labExp.typeCert.typeDer configure -text "[mc {DER-format}]"
    .saveCert.labExp.typeCert.typePem configure -text "[mc {PEM-format}]"
    .saveCert.labExp.but.butSave configure -text [mc "Export"]
    .saveCert.labExp.but.butCan configure -text [mc Exit]
    set nick [.saveCert.labExp.listCert get]
    set txt ".text.text"
    if {$::dercert != ""} {
	if {$::tekcert == "pkcs11"} {
	    $txt delete 0.0 end
	    if {[llength $::certs_p11] > 0} {
		::fileWithCert $txt $nick $::arrayCer($nick)
	    }
	} elseif {$::tekcert == "file"} {
	    set f [.saveCert.labExp.listCert get]
	    $txt delete 0.0 end
	    ::fileWithCert $txt "" $f
	}
    }
}

proc typeview {w} {
    global avecdet
    global loadfile
    if {$avecdet == 0} {
	set avecdet 1
	$w.tit.left.caption configure -image ::img::avec_det    
	$w.tit.left.caption configure -state normal    
	pack forget $w.text
	if {$::dercert != "" } {
	    pack $w.scroll -side right -fill y -padx {0 5}
	}
	pack $w.text -expand yes -fill both -side bottom
	$w.text configure -foreground #000000
	focus -force $w.text
    } else {
	set avecdet 0
	$w.tit.left.caption configure -image ::img::not_det   
	$w.tit.left.caption configure -state normal    
	$w.text configure -foreground #ffffff
	pack forget $w.scroll
	focus -force $w.text
    }

}

proc chainfromcert {cert dir} {
    global count
    if {$cert == "" } {
	exit
    }
    set asndata [cert_to_der $cert]
    if {$asndata == "" } {
#Файл содержит все что угодно, только не сертификат
	return -1
    }
    array set cert_parse [::pki::x509::parse_cert $asndata]
    array set extcert $cert_parse(extensions)
    if {![info exists extcert(1.3.6.1.5.5.7.1.1)]} {
#В сертификате нет расширений
	return 0
    }
    set a [lindex $extcert(1.3.6.1.5.5.7.1.1) 0]
#    if {$a == "false"} {
#	puts $a
#    }
#Читаем ASN1-последовательность расширения в Hex-кодировке
    set b [lindex $extcert(1.3.6.1.5.5.7.1.1) 1]
#Переводим в двоичную кодировку
    set c [binary format H* $b]
#Sequence 1.3.6.1.5.5.7.1.1
    ::asn::asnGetSequence c c_par_first
#Цикл перебора значений в засширении 1.3.6.1.5.5.7.1.1
    while {[string length $c_par_first] > 0 } {
#Выбираем очередную последовательность (sequence)
	::asn::asnGetSequence c_par_first c_par
#Выбираем oid из последовательности
	::asn::asnGetObjectIdentifier c_par c_type
	set tas1 [::pki::_oid_number_to_name $c_type]
#Выбираем установленное значение
	::asn::asnGetContext c_par c_par_two
#Ищем oid с адресом корневого сертификата
	if {$tas1 == "1.3.6.1.5.5.7.48.2" } {
#Читаем очередной корневой сертификат
	    set certca [readca $c_par $dir]
	    if {$certca == ""} {
#Прочитать сертификат не удалось. Ищем следующую точку с сертификатом
		continue
	    } else {
#Сохраняем корневой сертификат в указанном каталоге
		set f [file join $dir [file tail $c_par]]
		set fd [open $f w]
		chan configure $fd -translation binary
		puts -nonewline $fd $certca 
		close $fd
		lappend ::listcert $f
		incr count
#		puts "cert $count from $c_par"
#ПОДЫМАЕМСЯ по ЦЕПОЧКЕ СЕРТИФИКАТОВ ВВЕРХ
		chainfromcert $certca $dir
		continue
	    }
	} elseif {$tas1 == "1.3.6.1.5.5.7.48.1" } {
#	    puts "OCSP server (oid=$tas1)=$c_par"
	}
    }
# Цепочка закончилась
    return $count
}


proc readca {url dir} {
    set cer ""
#Проверяем тип протокола
    if { "https://" == [string range $url 0 7]} {
puts "HTTPS=$url"
	http::register https 443 ::tls::socket 
    }
#Читаем сертификат в бинарном виде
    if {[catch {set token [http::geturl $url -binary 1]
#получаем статус выполнения функции
	set ere [http::status $token]
	if {$ere == "ok"} {
#Получаем код возврата с которым был прочитан сертификат
	    set code [http::ncode $token]
	    if {$code == 200} {
#Сертификат успешно прочитан и будет созвращен
                set cer [http::data $token]
    	    } elseif {$code == 301 || $code == 302} {
#Сертификат перемещен в другое место, получаем его 
         		set newURL [dict get [http::meta $token] Location]
puts "newURL=$newURL"
#Читаем сертификат с другого сервера
          		set cer [readca $newURL $dir]
    	    } else {
#Сертификат не удалось прочитать
        	set cer ""
    	    }
        } 
    } error]} {
#Сертификат не удалось прочитать, нет узла в сети
	set cer ""
    }
    return $cer
}

proc ::removecert {} {
    global yespas
    global pass
    global nickTok
    puts "nickTok=$nickTok"
    if {[llength $::certs_p11] < 1} {
	return
    }

    set lists [listts $::handle]
    if {[llength $lists] == 0} {
	tk_messageBox -title [mc "PKCS#11 library"] -icon info -message "[mc {Tokens not present}]\n" 
	return
    }
    set find 0
    foreach {lab token_slotid} $lists {
	if {$lab == $nickTok } {
	    set find 1
	    break
	} 
    }
    if {$find == 0} {
	tk_messageBox -title [mc "PKCS#11 library"] -icon info -message "[mc {Token not find}]\n" 
	return
    }
#Ввод PIN-кода
    set yespas ""
    set pass ""
    set titpin "[mc {Token}]: $nickTok"
    wm title .topPinPw $titpin
    wm state .topPinPw normal
    wm state .topPinPw withdraw
    wm state .topPinPw normal
    raise .topPinPw
    grab .topPinPw
    focus .topPinPw.labFrPw.entryPw
    set yespas ""
    vwait yespas
    grab release .topPinPw
    if { $yespas == "no" } {
	return
    }
    set yesno "no"
    set uu [dict create pkcs11_handle $::handle]
    dict set uu pkcs11_slotid $token_slotid
    lappend uu "pkcs11_id"
    lappend uu $::pkcs11id
    set answer [tk_messageBox -icon question \
            -message "[mc {Remove keypair}]\n[mc {Yes/No}]?" \
            -title [mc "Remove"] \
            -type yesno]

#puts "Password=$pass"
    if {$answer != "yes"} {
	if { [pki::pkcs11::login $::handle $token_slotid $pass] == 0 } {
	    puts "Bad password"
	    return
	}
	pki::pkcs11::delete cert $uu
	pki::pkcs11::logout $::handle $token_slotid
    } else {
	if { [pki::pkcs11::login $::handle $token_slotid $pass] == 0 } {
	    puts "Bad password"
	    return
	}
	pki::pkcs11::delete all $uu
	pki::pkcs11::logout $::handle $token_slotid
    }
    unset pass
    updatetok
    return 
}

proc ::renamecert {} {
    global yespas
    global pass
    global nickTok
    if {[llength $::listx509] < 1} {

	return
    }
    puts "RENAMECERT nickTok=$nickTok"
#Ввод МЕТКИ
    set yespas ""
    set pass ""
    set titpin "[mc {Token}]: $nickTok"
    wm title .topPinPw $titpin
    wm state .topPinPw normal
    wm state .topPinPw withdraw
    wm state .topPinPw normal
    raise .topPinPw
    grab .topPinPw
.topPinPw.labFrPw configure -text "[mc {Enter label for this certificate}]"
pack forget .topPinPw.labFrPw.entryPw
pack .topPinPw.labFrPw.entryLb -fill x -expand 1 -padx 5 -pady 5  -ipady 2
    .topPinPw.labFrPw.entryLb insert end  [.saveCert.labExp.listCert get]

    focus .topPinPw.labFrPw.entryLb
    set yespas ""
    vwait yespas
    grab release .topPinPw
    pack forget .topPinPw.labFrPw.entryLb
    pack .topPinPw.labFrPw.entryPw -fill x -expand 1 -padx 5 -pady 5  -ipady 2
    .topPinPw.labFrPw.entryLb delete 0 end
    .topPinPw.labFrPw.entryPw delete 0 end
    .topPinPw.labFrPw.entryPw configure -show *
    .topPinPw.labFrPw configure -text "[mc {Enter the PIN code and press Enter}]"
    if { $yespas == "no" } {
	return
    }
    set newlab $pass
    set yesno "no"
    puts "RENAME=$::pkcs11id"
    puts "NEWLAB=$newlab"
#Ввод PIN-кода
    set yespas ""
    set pass ""
    set titpin "[mc {Token}]: $nickTok"
    wm title .topPinPw $titpin
    wm state .topPinPw normal
    wm state .topPinPw withdraw
    wm state .topPinPw normal
    raise .topPinPw
    grab .topPinPw
    focus .topPinPw.labFrPw.entryPw
    set yespas ""
    vwait yespas
    grab release .topPinPw
    if { $yespas == "no" } {
	    return
    }
    set yesno "no"
    puts "nickTok=$nickTok"
    set lists [listts $::handle]
    if {[llength $lists] == 0} {
	tk_messageBox -title [mc "Rename"] -icon info -message "[mc {Tokens not present}]\n" 
	return
    }

    set uu [dict create pkcs11_handle $::handle]
    dict set uu pkcs11_slotid $::slotid_tek
    lappend uu "pkcs11_id"
    lappend uu $::pkcs11id
    lappend uu "pkcs11_label"
    lappend uu $newlab
    if { [pki::pkcs11::login $::handle $::slotid_tek $pass] == 0 } {
	tk_messageBox -title [mc "Rename"] -icon info -message "[mc {Bad PIN}]"
	unset uu
	return
    }
#puts "PKCS11ID=$::pkcs11id"
#puts "PKCS11_LAB=$newlab"

    pki::pkcs11::rename all $uu
pki::pkcs11::logout $::handle $::slotid_tek
puts "Установлена метка $newlab"
    unset uu
    tk_messageBox -title [mc "Rename"] -icon info -message "[mc {Label certificate}] \n$::pkcs11id" -detail "$newlab"
    updatetok

}

proc ::verifysign {} {
    global pass
    global yespas
    global nickTok
    global home
    if {$::pkcs11_module == ""} {
	tk_messageBox -title [mc "Verify signature"] -icon error -message "[mc {Choose a PKCS#11 library for you Token}]"
	return -1
    }
    if {$::dercert == "" } {
	    tk_messageBox -title [mc "Verify signature"] -icon info -message [mc "No certificate selected"]
	    return -1
    }

    set cert_user $::dercert
    set caforver [::loadchain "ver"]
    if {$caforver == "" } {
	if {$::issuercert == $::subjectcert} {
	    set ms [mc "This is a root self-signed certificate"]
	    set dt [mc "The validity of a self-signed certificate is determined by the owner"]
	    tk_messageBox -title [mc "Verify signature"] -icon info -message $ms -detail $dt
	    set cert_CA $::dercert
	} else {
    	    set message "[mc {Cannot load CA}]\n[mc {You specify the file with the CA}]?"
    	    set answer [tk_messageBox -icon question \
                -message $message \
                -title [mc "Load CA"] \
                -detail [mc "Unable to verify certificate signature"] \
                -type yesno]

    	    if {$answer != "yes"} {
    		return 
    	    }
    	    set typeFile {
		{"[mc {Certificate to DER format}]"    .der}
		{"[mc {Certificate to DER format}]"    .cer}
		{"[mc {Certificate to PEM format}]"    .pem}
		{"[mc {Certificate to PEM format}]"    .crt}
		{"[mc {All}]"    *}
	    }
	    set typeFile [subst $typeFile]
	    set titleS [mc "Select file with certificate CA"]
	    set caforver [tk_getOpenFile -title $titleS -filetypes $typeFile -initialdir "$home"]
	    if {$caforver == ""} {
		return
	    }
	    set fd [open $caforver]
	    chan configure $fd -translation binary
	    set data [read $fd]
	    close $fd
	    set cert_CA [cert_to_der $data]
#	tk_messageBox -title [mc "Verify signature"] -icon info -message [mc "CAFORVER=$caforver"]	}
    } else {
	set fd [open $caforver]
	chan configure $fd -translation binary
	set data [read $fd]
	close $fd
	set cert_CA [cert_to_der $data]
#	tk_messageBox -title [mc "Verify signature"] -icon info -message [mc "CAFORVER from CHAIN=$caforver"]
    }
    array set cert_parse_user [::pki::x509::parse_cert $cert_user]
    array set cert_parse_CA [::pki::x509::parse_cert $cert_CA]
    if {![info exists cert_parse_CA]} {
	tk_messageBox -title [mc "Verify signature"] -icon error -message [mc "CA certificate bad=$caforver"]
	return
    }

    set aa [dict create pkcs11_handle $::handle pkcs11_slotid $::slotid_tek]
    set tbs_cert [binary format H* $cert_parse_user(cert)]
#puts "SIGN_ALGO1=$cert_parse(signature_algo)"
    catch {set signature_algo_number [::pki::_oid_name_to_number $cert_parse_user(signature_algo)]}
    if {![info exists signature_algo_number]} {
	set signature_algo_number $cert_parse_user(signature_algo)
    }
#puts "SIGN_ALGO=$signature_algo_number"
    switch -- $signature_algo_number {
	"1.2.643.2.2.3" - "1 2 643 2 2 3" { 
#    "GOST R 34.10-2001 with GOST R 34.11-94"
	    set digest_algo "gostr3411"
	}
	"1.2.643.7.1.1.3.2" - "1 2 643 7 1 1 3 2" {
#     "GOST R 34.10-2012-256 with GOSTR 34.11-2012-256"
	    set digest_algo "stribog256"
	}
	"1.2.643.7.1.1.3.3" - "1 2 643 7 1 1 3 3" { 
#    "GOST R 34.10-2012-512 with GOSTR 34.11-2012-512"
	    set digest_algo "stribog512"
	}
        default {
	    tk_messageBox -title [mc "Verify signature"] -icon info -message [mc "Cannot veridy signature\nUnknown signature algo"] -detail "$signature_algo_number"
	    return
	}
    }
#Проверка наличия механизмов
    set llmech [pki::pkcs11::listmechs $::handle $::slotid_tek]
    switch -- $digest_algo {
	    stribog256 {
		    set err [string first "0xD4321012" $llmech]
		}
	    stribog512 {
		    set err [string first "0xD4321013" $llmech]
		}	
	    gostr3411 {
		    set err [string first "0x1210" $llmech]
		}	
	    default {
		set err -1
	    }
    }
    if {$err == -1} {
	tk_messageBox -title [mc "Verify signature"] -icon info -message [mc "Unsupported algorithm/mechanism"] -detail "$digest_algo\n$signature_algo_number"
	return -1
    }

#Посчитать хэш от tbs-сертификата!!!!
    set digest_hex    [pki::pkcs11::digest $digest_algo $tbs_cert  $aa]
#Получаем asn-структуру публичного ключа
#Создаем список ключевых элементов

    binary scan $cert_CA H* cert_CA_hex
    array set infopk [pki::pkcs11::pubkeyinfo $cert_CA_hex  [list pkcs11_handle $::handle pkcs11_slotid $::slotid_tek]] 

    set lpk [dict create pkcs11_handle $::handle pkcs11_slotid $::slotid_tek]
#Добавляем pybkeyinfo в список ключевых элементов
    lappend lpk "pubkeyinfo"
    lappend lpk $infopk(pubkeyinfo)

#    array set lpkar $lpk
#    parray lpkar
#puts "Enter PIN user for you token \"$token_slotlabel\":"
set password "01234567"
#gets stdin password
#Ввод PIN-кода
    set yespas ""
    set pass ""
    set titpin "[mc {Token}]: $nickTok"
    wm title .topPinPw $titpin
    wm state .topPinPw normal
    wm state .topPinPw withdraw
    wm state .topPinPw normal
    raise .topPinPw
    grab .topPinPw
    focus .topPinPw.labFrPw.entryPw
    set yespas ""
    vwait yespas
    grab release .topPinPw
    if { $yespas == "no" } {
	return
    }
    set yesno "no"

    if { [pki::pkcs11::login $::handle $::slotid_tek $pass] == 0 } {
	pki::pkcs11::logout $::handle $::slotid_tek
	tk_messageBox -title [mc "Verify signature"] -icon info -message [mc "Cannot veridy signature\nBad PIN-code"] 
	return -1
    }
    pki::pkcs11::logout $::handle $::slotid_tek

    if {[catch {set verify [pki::pkcs11::verify $digest_hex $cert_parse_user(signature) $lpk]} res] } {
	tk_messageBox -title [mc "Verify signature"] -icon info -message [mc "Cannot veridy signature\n"] -detail $res 
	return -1
    }
    if {$verify != 1} {
	tk_messageBox -title [mc "Verify signature"] -icon info -message [mc "Certificate signature"] -detail [mc "Electronic signature bad"]
    } else {
	tk_messageBox -title [mc "Verify signature"] -icon info -message [mc "Certificate signature"] -detail [mc "Electronic signature mathematically correct"]
    }
    return $verify

}


proc ::loadchain {type} {
    global count
    global loadfile
    global home
    if {$::dercert == "" } {
	    tk_messageBox -title [mc "Load chain CA"] -icon info -message [mc "No certificate selected"]
	    return -1
    }
    set data $::dercert
    set dir [tk_chooseDirectory -initialdir $home -title [mc "Choos directory for chain"]]
    if {$dir == ""} {
	return ""
    }
    set ::listcert {}
    set count 0
    set depth [chainfromcert $data $dir]
#    puts "DEPTH=$depth"
#    puts $::listcert
    if {$depth == -1} {
	puts "Bad file with certificate=$file"
	return ""
    }
    set lenchain [llength $::listcert]
    set a ""
    set a "[mc {Files with CA certificates}]:\n\n"
    set caforver ""
    foreach b $::listcert {
	set caforver $b
	set a "$a  $b\n"
    }
    if {$type == "chain"} {
	tk_messageBox -title [mc "Load chain CA"] -icon info -message "[mc {Length chain CA}] - $lenchain" -detail $a
    }
    return [lindex $::listcert 0]
}

proc ::loadcrl {} {
    global count
    global loadfile
    global home
    if {$::dercert == "" } {
	    tk_messageBox -title [mc "Load CRL"] -icon info -message [mc "No certificate selected"]
	    return
    }
    set valid [cert_valid_date]
    puts "VALID=$valid"
    if {[lindex $valid 0] == 0} {
	set ms [mc [lindex $valid 1]]
        set answer [tk_messageBox -icon question \
                -message "$ms.\n[mc Continue]?" \
                -title [mc "Load CRL"] \
                -type yesno]

        if {$answer != "yes"} {
    	    return 
        }
    }
    if {$::crlfile == "" } {
	if {$::issuercert == $::subjectcert} {
	    set ms [mc "This is a root self-signed certificate"]
	    set dt [mc "The validity of a self-signed certificate is determined by the owner"]
	    tk_messageBox -title [mc "Load chain CA"] -icon info -message $ms -detail $dt
	    return
	}    
        set message "[mc {Cannot load CRL}]\n[mc {You specify the file with the CRL}]?"
        set answer [tk_messageBox -icon question \
                -message $message \
                -title [mc "Load CRL"] \
                -detail [mc "Unable to verify certificate validity"] \
                -type yesno]

        if {$answer != "yes"} {
    	    return 
        }
        set typeFile {
	    {"CRL"    .crl}
	    {"[mc {DER-format}]"    .cer}
	    {"[mc {PEM-format}]"    .pem}
	    {"[mc {All}]"    *}
	}
	set typeFile [subst $typeFile]
	set titleS [mc "Select file with CRL"]
	set filecrl [tk_getOpenFile -title $titleS -filetypes $typeFile -initialdir "$home"]
	if {$filecrl == ""} {
	    return
	}
	set f [open $filecrl r]
	chan configure $f -translation binary
	set crl [read $f]
	close $f

	array set valcert [validaty_cert_from_crl $crl $::sncert $::issuercert]
	if {[info exists valcert(error)]} {
	    tk_messageBox -title [mc "Load CRL"] -icon info -message "[mc {CRL from auther published}]\n$filecrl" -detail $valcert(issue)
	    return
	}
	if {![info exists valcert(revokeDate)]} {
#Сертификат не отозван
	    set valc "[mc {Certificate valid}] \(N=$::sncert\)"
	} else {
	    set t [string trimright $valcert(revokeDate) "Z"]
	    set year "20[string range $t 0 1]"
	    set month [string range $t 2 3]
	    set day [string range $t 4 5]
	    set hour [string range $t 6 7]
	    set minute [string range $t 8 9]
	    set second [string range $t 10 11]
#	puts  "$day $month $year $hour $minute"
	    set revdate "$day/$month/$year $hour:$minute:$second"
#	    set valc "[mc {Certificate revoked}] \(N=$valcert(sernumrev)\)"
	    set valc "[mc {Certificate revoked}] \(N=$valcert(sernumrev)\): $revdate"
	}
	set t [string trimright $valcert(publishDate) "Z"]
	set year "20[string range $t 0 1]"
	set month [string range $t 2 3]
	set day [string range $t 4 5]
	set hour [string range $t 6 7]
	set minute [string range $t 8 9]
	set second [string range $t 10 11]
	set publishdate "$day/$month/$year $hour:$minute:$second"
	set t [string trimright $valcert(nextDate) "Z"]
	set year "20[string range $t 0 1]"
	set month [string range $t 2 3]
	set day [string range $t 4 5]
	set hour [string range $t 6 7]
	set minute [string range $t 8 9]
	set second [string range $t 10 11]
	set nextdate "$day/$month/$year $hour:$minute:$second"
#    set nextdate [clock format $tt -format "%d/%m/%Y %R %Z"]

	set valc "$valc\n[mc {CRL release date}]: $publishdate\n[mc {Next CRL issue date}]: $nextdate"
	tk_messageBox -title [mc "Load CRL"] -icon info -message "[mc {Load CRL from}]\n$filecrl" -detail $valc
	return
    }
    set dir [tk_chooseDirectory -initialdir $home -title [mc "Select a directory to save the CRL"]]
    if {$dir == ""} {
	return
    }
    set filecrl ""
    set pointcrl ""
    foreach pointcrl $::crlfile {
	    set filecrl [readca $pointcrl $dir]
	    if {$filecrl != ""} {
		set f [file join $dir [file tail $pointcrl]]
		set fd [open $f w]
		chan configure $fd -translation binary
		puts -nonewline $fd $filecrl
		close $fd
		set filecrl $f
		break
	    } 
#Прочитать CRL не удалось. Берем следующую точку с CRL
    }
    if {$filecrl == ""} {
        set message "[mc {Cannot load CRL}]\n[mc {You specify the file with the CRL}]"
        set answer [tk_messageBox -icon question \
                -message $message \
                -title [mc "Load CRL"] \
                -detail [mc "Unable to verify certificate validity"] \
                -type yesno]

        if {$answer != "yes"} {
    	    return 
        }
        set typeFile {
	    {"CRL"    .crl}
	    {"[mc {DER-format}]"    .cer}
	    {"[mc {PEM-format}]"    .pem}
	    {"[mc {All}]"    *}
	}
	set typeFile [subst $typeFile]
	set titleS [mc "Select file with CRL"]
	set filecrl [tk_getOpenFile -title $titleS -filetypes $typeFile -initialdir "$home"]
	if {$filecrl == ""} {
	    return
	}
#	tk_messageBox -title [mc "Load CRL"] -icon error -message [mc "Cannot load CRL"] -detail [mc "Unable to verify certificate validity"]
#	return
    }

    set ss "[mc {File with CRL}]:\n$filecrl"
    set f [open $filecrl r]
    chan configure $f -translation binary
    set crl [read $f]
    close $f

#    set ::sncert 4604
    array set valcert [validaty_cert_from_crl $crl $::sncert $::issuercert]
    if {[info exists valcert(error)]} {
	tk_messageBox -title [mc "Load CRL"] -icon info -message "[mc {CRL from auther published}]\n$filecrl" -detail $valcert(issue)
	return
    }
#    parray valcert
    set revdate ""
    if {![info exists valcert(revokeDate)]} {
#Сертификат не отозван
	set valc "[mc {Certificate valid}] \(N=$::sncert\)"
    } else {
	set t [string trimright $valcert(revokeDate) "Z"]
	set year "20[string range $t 0 1]"
	set month [string range $t 2 3]
	set day [string range $t 4 5]
	set hour [string range $t 6 7]
	set minute [string range $t 8 9]
	set second [string range $t 10 11]
#	puts  "$day $month $year $hour $minute"
	set revdate "$day/$month/$year $hour:$minute:$second"
#	set revdate [clock format $tt -format "%d/%m/%Y %R %Z"]
	set valc "[mc {Certificate revoked}] \(N=$valcert(sernumrev)\): $revdate"
	if {[info exists valcert(reasone)]} {
	    set valc "$valc\n[mc {Revocation reason code}]: $valcert(reasoneData)"
	}
    }
    set t [string trimright $valcert(publishDate) "Z"]
	set year "20[string range $t 0 1]"
	set month [string range $t 2 3]
	set day [string range $t 4 5]
	set hour [string range $t 6 7]
	set minute [string range $t 8 9]
	set second [string range $t 10 11]
	set publishdate "$day/$month/$year $hour:$minute:$second"
    set t [string trimright $valcert(nextDate) "Z"]
	set year "20[string range $t 0 1]"
	set month [string range $t 2 3]
	set day [string range $t 4 5]
	set hour [string range $t 6 7]
	set minute [string range $t 8 9]
	set second [string range $t 10 11]
	set nextdate "$day/$month/$year $hour:$minute:$second"
#    set nextdate [clock format $tt -format "%d/%m/%Y %R %Z"]

    set valc "$valc\n[mc {CRL release date}]: $publishdate\n[mc {Next CRL issue date}]: $nextdate"
    tk_messageBox -title [mc "Load CRL"] -icon info -message "[mc {Load CRL from}]\n$pointcrl \n$ss" -detail $valc
}

proc ::pki::x509::parse_cert_pubkeyinfo {cert_hex} {
	array set ret [list]
	set wholething [binary format H* $cert_hex]
	::asn::asnGetSequence wholething cert

	::asn::asnPeekByte cert peek_tag
	if {$peek_tag != 0x02} {
		# Version number is optional, if missing assumed to be value of 0
		::asn::asnGetContext cert - asn_version
		::asn::asnGetInteger asn_version ret(version)
	}

	::asn::asnGetBigInteger cert ret(serial_number)
	::asn::asnGetSequence cert data_signature_algo_seq
		::asn::asnGetObjectIdentifier data_signature_algo_seq ret(data_signature_algo)
	::asn::asnGetSequence cert issuer
	::asn::asnGetSequence cert validity
		::asn::asnGetUTCTime validity ret(notBefore)
		::asn::asnGetUTCTime validity ret(notAfter)
	::asn::asnGetSequence cert subject
	::asn::asnGetSequence cert pubkeyinfo
#LISSI
	binary scan $pubkeyinfo H* ret(pubkeyinfo)
	return $ret(pubkeyinfo)
}


proc parse_key_gost {pubkeyinfo_hex} {
    array set ret [list]

    set pubkeyinfo [binary format H* $pubkeyinfo_hex]

    ::asn::asnGetSequence pubkeyinfo pubkey_algoid
	::asn::asnGetObjectIdentifier pubkey_algoid ret(pubkey_algo)
    ::asn::asnGetBitString pubkeyinfo pubkey
    
#		"1.2.643.2.2.19" -
#		"1.2.643.7.1.1.1.1" -
#		"1.2.643.7.1.1.1.2" 
#	gost2001, gost2012-256,gost2012-512
    set pubkey [binary format B* $pubkey]
    binary scan $pubkey H* ret(pubkey)
    ::asn::asnGetSequence pubkey_algoid pubalgost
#OID - параметра
    ::asn::asnGetObjectIdentifier pubalgost ret(paramkey)
#OID - Функция хэша
    ::asn::asnGetObjectIdentifier pubalgost ret(hashkey)
#puts "ret(paramkey)=$ret(paramkey)\n"
#puts "ret(hashkey)=$ret(hashkey)\n"
#parray ret
    return [array get ret]

}

proc issuerpol {iss_hex} {
    array set ret [list]
    set iss [binary format H* $iss_hex]
    ::asn::asnGetSequence iss iss_pol
    for {set i 1} {[string length $iss_pol] > 0}  {incr i} {
	::asn::asnGetUTF8String iss_pol ret(isspol$i)
    }
    return [array get ret]
}
proc subjectpol {iss_hex} {
    array set ret [list]
    set iss [binary format H* $iss_hex]
#    ::asn::asnGetSequence iss iss_pol
    for {set i 1} {[string length $iss] > 0}  {incr i} {
	::asn::asnGetUTF8String iss ret(isspol$i)
    }
    return [array get ret]
}

proc extku {ku_hex} {
    set typeku {"Digital signature" "Non-Repudiation" "Key encipherment" "Data encipherment" "Key agreement" "Certificate signature" "CRL signature" "Encipher Only" "Decipher Only"}
#    array set ret [list]
    set ku [binary format H* $ku_hex]
    ::asn::asnGetBitString ku ku_bin
    set ret {}
    for {set i 0} {$i < [string length $ku_bin]}  {incr i} {
#	puts "I=$i"
#	puts [lindex $typeku $i]
	if {[string range $ku_bin $i $i] > 0 } {
	    lappend ret [lindex $typeku $i]
	}
    }
#    puts $ret
    return $ret
}

proc extpol {pol_hex} {
    set pol [binary format H* $pol_hex]
    set ret {}
    ::asn::asnGetSequence pol oid_pol
    for {set i 1} {[string length $oid_pol] > 0}  {incr i} {
	::asn::asnGetSequence oid_pol oid_pol1
	::asn::asnGetObjectIdentifier oid_pol1 ret1
	lappend ret $ret1
    }
#    puts $ret
    return $ret
}

proc extkeyuse {keyuse_hex} {
    set use [binary format H* $keyuse_hex]
    set ret {}
    ::asn::asnGetSequence use oid_use
    for {set i 1} {[string length $oid_use] > 0}  {incr i} {
	::asn::asnGetObjectIdentifier oid_use ret1
	lappend ret $ret1
    }
#    puts "EXTKEYUSE=$ret"
    return $ret
}

proc chainocsp {chain_hex} {
    set chain [binary format H* $chain_hex]
    set ret {}
    ::asn::asnGetSequence chain c_par_first
    while {[string length $c_par_first] > 0 } {
#Выбираем очередную последовательность (sequence)
	::asn::asnGetSequence c_par_first c_par
#Выбираем oid из последовательности
	::asn::asnGetObjectIdentifier c_par c_type
	set tas1 [::pki::_oid_number_to_name $c_type]
#Выбираем установленное значение
	::asn::asnGetContext c_par c_par_two
#Ищем oid с адресом корневого сертификата
	if {$tas1 == "1.3.6.1.5.5.7.48.2" } {
#Читаем очередной корневой сертификат
	    lappend ret "CA Issuers"
	    lappend ret $c_par
#	    puts "CA (oid=$tas1)=$c_par"
	} elseif {$tas1 == "1.3.6.1.5.5.7.48.1" } {
	    lappend ret "OCSP"
	    lappend ret $c_par
#	    puts "OCSP server (oid=$tas1)=$c_par"
	}
    }
# Цепочка закончилась
    return $ret
}

proc crlpoints {crl_hex} {
    set crl [binary format H* $crl_hex]
    set ret {}
    ::asn::asnGetSequence crl p_crl_first
    set lencontspec 4
    while {[string length $p_crl_first] > 0 } {
#Выбираем очередную последовательность (sequence)
	::asn::asnGetSequence p_crl_first c_par
#Пропускаем contextspecific - 0xA0
	::asn::asnGetBytes c_par $lencontspec context_0xa0
	::asn::asnGetContext c_par ux
#	puts $c_par
	lappend ret $c_par
    }
    return $ret
}

proc altname {alt_hex} {
    set ret {}
    set listname [binary format H* $alt_hex]
    ::asn::asnGetSequence listname names
    while {[string length $names] > 0 } {
	::asn::asnGetByte names tag
	if {$tag == 0x82 || $tag == 0x81} {
#0x82 DNS Name ; 0x81 - RFC 822
#	    puts "0x82"
########################
	    ::asn::asnGetByte names taglen
	    set len82 [format "%i" $taglen]
#	    puts $len82
	    incr len82 -1
	    if {$tag == 0x82 } {
#Читаем очередной корневой сертификат
		lappend ret "DNS Name"
	    } elseif {$tag == 0x81 } {
		lappend ret "RFC822 Name"
	    } else {
		lappend ret [forma "0x%2x" $tag]
	    }
	    lappend ret [string range $names 0 $len82]
#	    puts $ret
	    incr len82
	    set a [string range $names $len82 end]
	    set names $a
#	    ::asn::asnGetByte validaty tag
	} 
    }
    return $ret
}

proc keyperiod {per_hex} {
    set per [binary format H* $per_hex]
    array set ret [list]
    ::asn::asnGetSequence per validaty
    ::asn::asnGetByte validaty tag
    if {$tag == 0x80} {
	::asn::asnGetByte validaty tag
	set len80 [format "%i" $tag]
	incr len80 -1
	set ret(notBefore) [string range $validaty 0 $len80]
	incr len80
	set a [string range $validaty $len80 end]
	set validaty $a
	::asn::asnGetByte validaty tag
    } else {
	set ret(notBefore) ""
    }
    if {$tag != 0x81} {
	set ret(After) ""
	return [array get ret]
    }
    ::asn::asnGetByte validaty tag
    set len81 [format "%i" $tag]
    incr len81 -1
    set ret(notAfter) [string range $validaty 0 $len81]
    incr len81
    return [array get ret]
}


proc autkeyid {autkey_hex} {
    set autkey [binary format H* $autkey_hex]
    array set ret [list]
#########################################
    set autkey [binary format H* $autkey_hex]
    array set ret [list]
    ::asn::asnGetSequence autkey fullkey
#Чтение KEY ID
    ::asn::asnGetContext fullkey - cont
#    puts "LEN=[string length $fullkey]"
    binary scan $cont H* ret(authKeyID)
    while {[string length $fullkey]} {
	::asn::asnGetByte fullkey tag1
	::asn::asnGetLength fullkey tag
	set len80 [format "0x%x" $tag]
#	if {$tag1 == 0xA4} {}
	set ee [string range $fullkey 0 0]
	if { $ee == 0} {
	    ::asn::asnGetSequence fullkey autdn
	    set ret(issuer) [::pki::x509::_dn_to_string $autdn]
	    ::asn::asnGetContext fullkey - sernum
	    binary scan $sernum H* ret(sernum)
	    break
	}
#	{}
    }

    return [array get ret]
}

proc cert_to_der {data} {
    if {[string first "-----BEGIN CERTIFICATE-----" $data] != -1} {
#	set data [string map {"\015\012" "\n"} $data]
	set data [string map {"\r\n" "\n"} $data]
    }
    array set parsed_cert [::pki::_parse_pem $data "-----BEGIN CERTIFICATE-----" "-----END CERTIFICATE-----"]
    set asnblock $parsed_cert(data)
    return $asnblock
}
proc del_comma {ldn} {
    set lret {}
    set ff ""
    foreach el $ldn {
	if {[string first "=" $el] != -1} {
	    if {$ff != ""} {
		lappend lret $ff
	    }
	    set ff $el
	} else {
	    set ff "$ff,$el"
	}
    }
    lappend lret $ff
    return $lret
}

proc ::fingerPrintCert {cert_hex signature_hex data_signature_algo} {
	set cert [binary format H* $cert_hex]
	set signature [binary format H* $signature_hex]
	binary scan $signature B* signature_bitstring
	set cert [::asn::asnSequence \
		$cert \
		[::asn::asnSequence \
			[::asn::asnObjectIdentifier [::pki::_oid_name_to_number $data_signature_algo]] \
			[::asn::asnNull] \
		] \
		[::asn::asnBitString $signature_bitstring] \
	]
	set lfin {}
	lappend lfin [::sha1::sha1 $cert]
	lappend lfin [::sha2::sha256 $cert]
	set ::dercert $cert
	return $lfin
}

proc ::fileWithCert {w nick cert_hex} {
    global avecdet
    global home
    global loadfile
    global dn_fields
    set ::dercert ""
    set ::parsecert ""
    $w delete 0.0 end
#    set loadfile ""
#Файл с сертификатм из командной строки
    set comline 1
 ###############
    array set cert_parse []
#    if {$nick == "" && $cert_hex == ""} {}
    if {$nick == "" } {
	unset -nocomplain cert_parse

        set typeFile {
	    {"[mc {Certificate to DER format}]"    .der}
	    {"[mc {Certificate to DER format}]"    .cer}
	    {"[mc {Certificate to PEM format}]"    .pem}
	    {"[mc {Certificate to PEM format}]"    .crt}
	    {"[mc {All}]"    *}
	}
#	puts "LOADFILE=$loadfile"
	set lastdir $home
	if {$loadfile != "" } {
		set lastdir [file dirname $loadfile]
	}
#	puts "LASTDIR=$lastdir"
	set titleS [mc "Select file with certificate"]
	if {$cert_hex != ""} {
	    set file $cert_hex
	} else {
	    set typeFile [subst $typeFile]
	    set file [tk_getOpenFile -title $titleS -filetypes $typeFile -initialdir "$lastdir"]
	}
	if {$file == ""} {
	    return
	}
	set fd [open $file]
	chan configure $fd -translation binary
	set data [read $fd]
	close $fd
	set asndata [cert_to_der $data]
	if {$asndata == "" } {
	    tk_messageBox -title [mc "Load certificate"] -icon error -message "$file" -detail [mc "The selected file does not contain a certificate"]
	    return
	}
	if {[catch {array set cert_parse [::pki::x509::parse_cert $asndata]} rc]} {
	    tk_messageBox -title [mc "Load certificate"] -icon error -message "$file" -detail [mc "The selected file does not contain a certificate"]
	    return
	}
	set loadfile $file
	set cert_hex $cert_parse(cert)
	set cert_parse(pubkeyinfo) [::pki::x509::parse_cert_pubkeyinfo $cert_parse(cert)]
	.saveCert.labExp.listCert configure -state normal
	.saveCert.labExp.listCert delete 0 end
	.saveCert.labExp.listCert insert end $file
	.saveCert.labExp.listCert configure -state readonly
	set ::tekcert "file"
	.saveCert.labExp.but.butSave configure -text [mc "Import"]
    } else {
 
############
	foreach certinfo_list $::certs_p11 {
	    unset -nocomplain cert_parse
	    array set cert_parse $certinfo_list
	    if {$cert_parse(pkcs11_label) == $nick} {
#Читаем публичный ключ
		set cert_parse(pubkeyinfo) [::pki::x509::parse_cert_pubkeyinfo $cert_parse(cert)]
		break
	    }
	}
	.saveCert.labExp.but.butSave configure -text [mc "Export"]
	set ::tekcert "pkcs11"
    }
#parray cert_parse
#    puts  $cert_parse(subject)
    set ::subjectcert "$cert_parse(subject)"
    set lsub [split $cert_parse(subject) ","]
    set lsub [del_comma $lsub]
#    puts $lsub

#    $w delete 0.0 end
    $w insert end [mc "Subject Name"] bold
    $w insert end "\n"
    foreach a $lsub {
	set ind [string first "=" $a]
	if {$ind == -1 } { continue }
	
	set oidsub [string trim [string range $a 0 $ind-1]]
	if {[info exists dn_fields($oidsub)]} {
	    set nameoid " ([mc $dn_fields($oidsub)])"
	} else {
	    set nameoid ""
	}

#	puts $nameoid
	set oidval "[mc [string trim [string range $a $ind+1 end]]]"
	if {$oidsub == "CN"} {
	    set ::cn_subject  $oidval
	}

	set oidsub "$oidsub$nameoid"
	$w insert end "\t$oidsub\t$oidval\n"  margins1
    }
    $w insert end [mc "Issuer Name"] bold
    $w insert end "\n"
#    puts  $cert_parse(issuer)
    set ::issuercert "$cert_parse(issuer)"
    set liss [split $cert_parse(issuer) ","]
    set liss [del_comma $liss]
#    puts $liss

    foreach a $liss {
	set ind [string first "=" $a]
	if {$ind == -1 } { continue }
	set oidsub [string trim [string range $a 0 $ind-1]]
	if {[info exists dn_fields($oidsub)]} {
	    set nameoid " ([mc $dn_fields($oidsub)])"
	} else {
	    set nameoid ""
	}

	set oidval "[mc [string trim [string range $a $ind+1 end]]]"
	if {$oidsub == "CN"} {
	    set ::cn_issuer  $oidval
	}

	set oidsub "$oidsub$nameoid"
	$w insert end "\t$oidsub\t$oidval\n"  margins1
    }
    $w insert end [mc "Issued Certificate"] bold
    $w insert end "\n"
    $w insert end "\t[mc "Version"]:\t$cert_parse(version)\n"  margins1
    set ::sncert "$cert_parse(serial_number)"
    $w insert end "\t[mc "Serial Number"]:\t$cert_parse(serial_number)\n"  margins1
    set ::notafter  $cert_parse(notAfter)
    set t $cert_parse(notAfter)
    set notafter [clock format $t -format "%d/%m/%Y %R %Z"]
    set ::notbefore $cert_parse(notBefore)
    set t $cert_parse(notBefore)
    set notbefore [clock format $t -format "%d/%m/%Y %R %Z"]
    $w insert end "\t[mc "Not Valid Before"]:\t$notbefore\n"  margins1
    $w insert end "\t[mc "Not Valid After"]:\t$notafter\n"  margins1
    set ver [mc "Expires"]
    $w insert end [mc "Certificate Fingerprints"] bold
    $w insert end "\n"
    set lfinger {}
    set lfinger [fingerPrintCert $cert_hex $cert_parse(signature) $cert_parse(signature_algo)]

    set fingerprint_sha1 [lindex $lfinger 0]
    set fingerprint_sha256 [lindex $lfinger 1]

    $w insert end "\t[mc "SHA1"]:\t[edithex $fingerprint_sha1]\n"  margins11
    $w insert end "\t[mc "SHA256"]:\t[edithex $fingerprint_sha256]\n"  margins11
    $w insert end [mc "Public Key Info"] bold
    $w insert end "\n"
    if {[string range $cert_parse(pubkey_algo) 0 7] == "1.2.643." || [string range $cert_parse(pubkey_algo) 0 7] == "ГОСТ Р 3" || [string range $cert_parse(pubkey_algo) 0 7] == "GOST R 3"} {
	$w insert end "\t[mc "Key Algorithm"]:\t[mc $cert_parse(pubkey_algo)]\n"  margins1
	$w insert end "\t[mc "Key Parameters"]:\n"  margins1
	array set ret [parse_key_gost $cert_parse(pubkeyinfo)]
#	parray ret
	$w insert end "\t[mc "sign param"]:\t[mc $ret(paramkey)]\n"  margins2
	$w insert end "\t[mc "hash param"]:\t[mc $ret(hashkey)]\n"  margins2
	set sek 4
	if {[string range $ret(pubkey) 2 3] != 40} {
	    set sek 6
	}
	set pk [edithex [string range $ret(pubkey) $sek end]]
	$w insert end "\t[mc "Public Key"]:\t$pk\n"  margins11
if {0} {
	$w insert end "\t[mc "Public Key"]:\t[string range $pk 0 23]\n"  margins11
	for {set i 24 } { $i < [string length $pk] } {incr i +24} {
	    $w insert end "\t\t[mc [string range $pk $i [expr {$i + 23}]]]\n"  margins1
	}
}
#Идентификатор ключа получателя
	set ::pkcs11id [::sha1::sha1 [binary format H* $ret(pubkey)]]

#?????	$w insert end "\t[mc "Key SHA1 Fingerprint"]:\t[edithex $pk_sha1]\n"  margins11
	
    } else {
	if {[string range $cert_parse(pubkey_algo) 0 2] == "rsa" } {
	    set ::pkcs11id [::sha1::sha1 [binary format H* $ret(pubkey)]]
	    $w insert end "\t[mc "Key Algorithm"]:\tRSA\n"  margins1
	    $w insert end "\t[mc "Key Size"]:\t$cert_parse(l)\n"  margins2
	    $w insert end "\t[mc "Public Key"]:\t[edithex $cert_parse(pubkey)]\n"  margins11
	} else {
	    $w insert end "\t[mc "Key Algorithm"]:\t$cert_parse(pubkey_algo)\n"  margins1
	    $w insert end "\t[mc "Key Info"]:\t[edithex $cert_parse(pubkeyinfo)]\n" margins11
	}
    }
    array set extcert $cert_parse(extensions)
#    parray extcert
    if {[info exists extcert(id-ce-basicConstraints)]} {
	$w insert end [mc "Basic Constraints"] bold
	$w insert end "\n"
	set basic $extcert(id-ce-basicConstraints)
#	puts $basic
	if {[lindex $basic 1] == 1} {
	    set typecert [mc "Yes"]
	} else {
	    set typecert [mc "No"]
	}
	$w insert end "\t[mc "Certificate Authority"]:\t$typecert\n"  margins1
	if {[lindex $basic 2] == -1} {
	    set lencert [mc "Unlimited"]
	} else {
	    set lencert [lindex $basic 2]
	}
	$w insert end "\t[mc "Max Path Length"]:\t$lencert\n"  margins1
	if {[lindex $basic 0] == 1} {
	    set critcert [mc "Yes"]
	} else {
	    set critcert [mc "No"]
	}
	$w insert end "\t[mc "Critical"]:\t$critcert\n"  margins1
	unset extcert(id-ce-basicConstraints)
    }
#  1 false -1
# первое поле критичность 1 - Да, false - нет    
# второе поле УЦ 1 - Да, false - нет
#Третье поле - длина пути : -1 - неограниченный, или значение 0 и т.д.
    if {[info exists extcert(1.2.643.100.112)]} {
#issuerSignTools
	array set pol [issuerpol [lindex $extcert(1.2.643.100.112) 1]]
	$w insert end [mc "issuerSignTool"] bold
	$w insert end "\n"
	$w insert end "\t[mc "Name CKZI"]:\t$pol(isspol1)\n"  margins1
	$w insert end "\t[mc "Name CA"]:\t$pol(isspol2)\n"  margins1
	$w insert end "\t[mc "Certificate SKZI CA"]:\t$pol(isspol3)\n"  margins1
	$w insert end "\t[mc "Certificate CA"]:\t$pol(isspol4)\n"  margins1
#	parray pol
	unset extcert(1.2.643.100.112)
    }
    if {[info exists extcert(1.2.643.100.111)]} {
#subjectSignTools
	array set pol [subjectpol [lindex $extcert(1.2.643.100.111) 1]]
	$w insert end [mc "subjectSignTool"] bold
	$w insert end "\n"
	$w insert end "\t[mc "User CKZI"]:\t$pol(isspol1)\n"  margins1
#	parray pol
	unset extcert(1.2.643.100.111)
    }
    if {[info exists extcert(id-ce-keyUsage)]} {
	$w insert end [mc "Key Usage"] bold
	$w insert end "\n"

	set ku [extku [lindex $extcert(id-ce-keyUsage) 1]]
	$w insert end "\t[mc "Usages"]:\t[mc [lindex $ku 0]]\n"  margins1
	for {set i 1 } { $i < [llength $ku] } {incr i} {
	    $w insert end "\t\t[mc [lindex $ku $i]]\n"  margins1
	}
#	puts $ku
	if {[lindex $extcert(id-ce-keyUsage) 0] == 1} {
	    set critcert [mc "Yes"]
	} else {
	    set critcert [mc "No"]
	}
	$w insert end "\t[mc "Critical"]:\t$critcert\n"  margins1
	unset extcert(id-ce-keyUsage)
    }
    if {[info exists extcert(id-ce-certificatePolicies)]} {
	$w insert end [mc "Certificate Policies"] bold
	$w insert end "\n"
	set lpol [extpol [lindex $extcert(id-ce-certificatePolicies) 1]]
	$w insert end "\t[mc "Policy Name"]:\t[mc [::pki::_oid_number_to_name [lindex $lpol 0]]]\n"  margins1
	for {set i 1 } { $i < [llength $lpol] } {incr i} {
	    $w insert end "\t\t[mc [::pki::_oid_number_to_name [lindex $lpol $i]]]\n"  margins1
	}
#	puts $ku
	if {[lindex $extcert(id-ce-certificatePolicies) 0] == 1} {
	    set critcert [mc "Yes"]
	} else {
	    set critcert [mc "No"]
	}
	$w insert end "\t[mc "Critical"]:\t$critcert\n"  margins1
	unset extcert(id-ce-certificatePolicies)
    }
    if {[info exists extcert(id-ce-subjectKeyIdentifier)]} {
	$w insert end [mc "Subject Key Identifier"] bold
	$w insert end "\n"
	$w insert end "\t[mc "Key ID"]:\t[edithex [string range [lindex $extcert(id-ce-subjectKeyIdentifier) 1] 4 end]]\n"  margins1
#	set ::pkcs11id [string range [lindex $extcert(id-ce-subjectKeyIdentifier) 1] 4 end]
	unset extcert(id-ce-subjectKeyIdentifier)
    }
    if {[info exists extcert(id-ce-privateKeyUsagePeriod) ]} {
	$w insert end [mc "Key Usage Period"] bold
	$w insert end "\n"
	array set keyperiod [keyperiod [lindex $extcert(id-ce-privateKeyUsagePeriod) 1]]
#	parray keyperiod
	set t $keyperiod(notBefore)
#	set notafter [clock format [clock scan "[string range $t 0 1]-[string range $t 2 3]-[string range $t 4 5] [string range $t 6 7]:[string range $t 8 9]"] -format "%d/%m/%Y %H:%M"]
	set year [string range $t 0 3]
	set month [string range $t 4 5]
	set day [string range $t 6 7]
	set hour [string range $t 8 9]
	set minute [string range $t 10 11]

#	puts  "$day $month $year $hour $minute"
	set notbefore "$day/$month/$year $hour:$minute"

	set t $keyperiod(notAfter)
	set notafter "[string range $t 6 7]/[string range $t 4 5]/[string range $t 0 3] [string range $t 8 9]:[string range $t 10 11]"
	$w insert end "\t[mc "Not Valid Before"]:\t$notbefore\n"  margins1
	$w insert end "\t[mc "Not Valid After"]:\t$notafter\n"  margins1

	if {[lindex $extcert(id-ce-privateKeyUsagePeriod) 0] == 1} {
	    set critcert [mc "Yes"]
	} else {
	    set critcert [mc "No"]
	}
	$w insert end "\t[mc "Critical"]:\t$critcert\n"  margins1
	unset extcert(id-ce-privateKeyUsagePeriod)
    }
    if {[info exists extcert(id-ce-authorityKeyIdentifier) ]} {
	$w insert end [mc "Certificate Authority Key Identifier"] bold
	$w insert end "\n"
	array set autkey [autkeyid [lindex $extcert(id-ce-authorityKeyIdentifier) 1]]
	$w insert end "\t[mc "Key ID"]:\t[edithex $autkey(authKeyID)]\n"  margins1
	if {[info exists autkey(issuer) ] } {
	    $w insert end "\t[mc "Directory Name"]:\t$autkey(issuer)\n"  margins1
	}
	if {[info exists autkey(sernum) ]} {
	    $w insert end "\t[mc "Serial Number"]:\t[edithex $autkey(sernum)]\n"  margins1
	}
	unset extcert(id-ce-authorityKeyIdentifier)
    }
    if {[info exists extcert(2.5.29.37) ]} {
	$w insert end [mc "Extended Key Usage"] bold
	$w insert end "\n"
	set listusage [extkeyuse [lindex $extcert(2.5.29.37) 1]]
	set oidt [string map {" " "."} [lindex $listusage 0]]
	if {[info exists ::payoid($oidt) ]} {
	    set poid " ($::payoid($oidt))"
	} else {
	    set poid ""
	}
#	$w insert end "\t[mc "Allowed Purposes"]:\t[mc [::pki::_oid_number_to_name [lindex $listusage 0]]]\n"  margins1
	$w insert end "\t[mc "Allowed Purposes"]:\t$oidt$poid\n"  margins1
	for {set i 1 } { $i < [llength $listusage] } {incr i} {
	    set oidt [string map {" " "."} [lindex $listusage $i]]
	    if {[info exists ::payoid($oidt) ]} {
		set poid " ($::payoid($oidt))"
	    } else {
		set poid ""
	    }
	    $w insert end "\t\t$oidt$poid\n"  margins1
	
#	    $w insert end "\t\t[mc [::pki::_oid_number_to_name [lindex $listusage $i]]]\n"  margins1
	}
	if {[lindex $extcert(2.5.29.37) 0] == 1} {
	    set critcert [mc "Yes"]
	} else {
	    set critcert [mc "No"]
	}
	$w insert end "\t[mc "Critical"]:\t$critcert\n"  margins1
	unset extcert(2.5.29.37)
    } 
    set ::chaincert ""
    if {[info exists extcert(1.3.6.1.5.5.7.1.1)]} {
	$w insert end [mc "Authority information Accesss"] bold
	$w insert end "\n"
	set listchain [chainocsp [lindex $extcert(1.3.6.1.5.5.7.1.1) 1]]
#	puts $listchain
	foreach {a b} $listchain {
	    $w insert end "\t[mc $a]:\tURI:$b\n"  margins1
	}
	set ::chaincert [lindex $extcert(1.3.6.1.5.5.7.1.1) 1]
#	puts "CHAIN=[lindex $extcert(1.3.6.1.5.5.7.1.1) 1]"
	unset extcert(1.3.6.1.5.5.7.1.1)
    } 

#    if {[info exists extcert(id-ce-cRLDistributionPoints)]} {}
    set ::crlfile ""
    if {[info exists extcert(2.5.29.31)]} {
	$w insert end [mc "CRL Distribution Points"] bold
	$w insert end "\n"
#	puts "CRL=$extcert(2.5.29.31)"
	set listcrl [crlpoints [lindex $extcert(2.5.29.31) 1]]
#	puts $listcrl
	foreach {crlp} $listcrl {
	    $w insert end "\t[mc "Distribution Point"]:\tURI:$crlp\n"  margins1
	}
#	set ::crlfile [lindex $extcert(2.5.29.31) 1]
	set ::crlfile  $listcrl
	unset extcert(2.5.29.31)
    }    
#extcert(id-ce-issuerAltName)          = false 3000
#extcert(id-ce-subjectAltName) 
    if {[info exists extcert(id-ce-issuerAltName)]} {
	$w insert end [mc "Issuer Alt Name"] bold
	$w insert end "\n"
#	puts "ALT ISSUER=$extcert(id-ce-issuerAltName)"
	set listalt [altname [lindex $extcert(id-ce-issuerAltName) 1]]
	foreach {a b} $listalt {
	    $w insert end "\t[mc $a]:\tURI:$b\n"  margins1
	}
	unset extcert(id-ce-issuerAltName)
    }
    if {[info exists extcert(id-ce-subjectAltName)]} {
	$w insert end [mc "Subject Alt Name"] bold
	$w insert end "\n"
#	puts "ALT=$extcert(id-ce-subjectAltName)"
	set listalt [altname [lindex $extcert(id-ce-subjectAltName) 1]]
	foreach {a b} $listalt {
	    $w insert end "\t[mc $a]:\tURI:$b\n"  margins1
	}
	unset extcert(id-ce-subjectAltName)
    }
    
    set listext [array get extcert]
    foreach {a b} $listext {
	$w insert end [mc "Extension"] bold
	$w insert end "\n"
	$w insert end "\t[mc "Identifier"]:\t$a\n"  margins1
	$w insert end "\t[mc "Value"]:\t[edithex [lindex $b 1]]\n"  margins1
	if {[lindex $b 0] == 1} {
	    set critcert [mc "Yes"]
	} else {
	    set critcert [mc "No"]
	}
	$w insert end "\t[mc "Critical"]:\t$critcert\n"  margins1
    }

    $w insert end [mc "Signature"] bold
    $w insert end "\n"
    $w insert end "\t[mc "Signature Algorithm"]\t[mc "$cert_parse(signature_algo)"]\n"  margins1
    $w insert end "\t[mc "Signature"]:\t[edithex $cert_parse(signature)]\n"  margins11
if {0} {
puts $cert_parse(signature)
    set pk [edithex $cert_parse(signature)]
    $w insert end "\t[mc "Signature"]:\t[string range $pk 0 23]\n"  margins11
    for {set i 24 } { $i < [string length $pk] } {incr i +24} {
	$w insert end "\t\t[mc [string range $pk $i [expr {$i + 23}]]]\n"  margins1
    }
}
    if {$avecdet == 0} {
	typeview .text
    }
    set ::parsecert [$w get 1.0 end]
}


proc exportparsecert {w} {
    global home
#puts "CERT=$::parsecert"
    if {[string length $::parsecert] < 2} {
	tk_messageBox -title [mc "Save certificate"] -icon info -message [mc "No certificate selected"]
	return
    }
    set typef {
	{"[mc {TXT-format}]"    *.txt}
	{"[mc {All}]"  *}
    }
    set typef [subst $typef]
    set savefile [tk_getSaveFile -title [mc "Save certificate"] -filetypes $typef -initialdir $home ]
    if {$savefile == ""} {
	return
    }
    if {[catch {set fd [open "$savefile" w]} rc]} {
	tk_messageBox -title [mc "Save certificate"] -icon error -message "Cannot open $savefile" -detail "$rc"
	return
    }
#    chan configure $fd -translation binary
    puts -nonewline $fd $::parsecert
    close $fd
    tk_messageBox -title [mc "Save certificate"] -icon info -message "$savefile" -detail [mc "File with parsed certificate"]
    return
}

proc mechTok {} {
	aboutUtil .about 1
	set llmech [pki::pkcs11::listmechs $::handle $::slotid_tek]
	foreach mech $llmech {
	    $w.text insert end $mech
	    $w.text insert end "\n"
	}
}

proc showContextMenu {w x y rootx rooty} {
    catch {destroy .contextMenu}
    menu .contextMenu -tearoff false
    .contextMenu add command -label [mc "Save certificate"] -command [list exportparsecert $w]
    .contextMenu configure -title "Certificate"
    .contextMenu add command -label [mc "Load chain CA"] -command {::loadchain "chain"}
    .contextMenu add command -label [mc "Validaty by CRL"] -command {::loadcrl}
    .contextMenu add command -label [mc "OCSP validity"] -command {return}
    if {$::tekcert == "pkcs11"} {
	.contextMenu add command -label [mc "Remove"] -command {::removecert}
	.contextMenu add command -label [mc "Rename"] -command {::renamecert}
    }
    if {$::pkcs11_module != ""} {
	.contextMenu add command -label [mc "Token mechanisms"] -command {aboutUtil .about 1}
    }
    .contextMenu add separator
    .contextMenu add command -label [mc "About"] -command {aboutUtil .about 0}

    tk_popup .contextMenu $rootx $rooty
    .contextMenu configure -activebackground #39b5da
    .contextMenu configure -background #e0e0da
}

proc updatetok {} {
puts "UpdateListTok"
    if {$::handle != ""} {
	set ::handle [pki::pkcs11::unloadmodule $::handle]
    }
    if {$::pkcs11_module == ""} {
	tk_messageBox -title [mc "PKCS#11 library"] -icon error -message "[mc {Choose a PKCS#11 library for you Token}]"
	return    
    }
#		set ::pkcs11_module "[encoding convertto cp1251 $file]"

    if {[catch {set ::handle [pki::pkcs11::loadmodule "$::pkcs11_module"]} result]} {
	set ::handle ""
	tk_messageBox -title [mc "PKCS#11 library"] -icon error -message "[mc {Bad PKCS#11 library}]\n$::pkcs11_module" -detail "$result\nlen=[string length $::pkcs11_module]"
	.fr3.ent configure -state normal
	.fr3.ent delete 0 end
	.fr3.ent configure -state readonly
	set ::pkcs11_module ""
	return
    }
#Список найденных токенов в слотах и сертификатов
    set ::slotid_tek ""
		set lists [listts $::handle]
		if {[llength $lists] == 0} {
		    .text.text delete 0.0 end
		    .saveCert.labExp.listCert configure -state normal
		    .saveCert.labExp.listCert delete 0 end
		    .saveCert.labExp.listCert configure -state readonly
		    .selecttok.tok configure -state normal
		    .selecttok.tok delete 0 end
		    .selecttok.tok configure -state readonly
		    tk_messageBox -title [mc "PKCS#11 library"] -icon info -message "[mc {Tokens not present}]\n" 
#Автоматическое обновление токенов
after 5000 updatetok
		    return
		}

		set i 0
		set ::listtok {}
		foreach {lab slotid} $lists {
		    puts "Токен \"$lab\" находится в слоте \"$slotid\""
		    lappend ::listtok $lab
		    incr i
		    if {$::slotid_tek == ""} {
			set ::slotid_tek $slotid
		    }
		}
		.selecttok.tok configure -values $::listtok
		.selecttok.tok configure -state normal
		.selecttok.tok delete 0 end
		.selecttok.tok insert end [lindex $::listtok 0]
		.selecttok.tok configure -state readonly
		puts "CountSlot=$i, $::listtok"
		puts "LEN_lists=[llength $lists]"
		puts $lists
		.saveCert.labExp.listCert configure -state normal
		set ::listx509 [listcerttok $::handle [lindex $lists 0] [lindex $lists 1]]
		.saveCert.labExp.listCert configure -values $::listx509
		.saveCert.labExp.listCert delete 0 end
		.saveCert.labExp.listCert insert end [lindex $::listx509 0]
		.saveCert.labExp.listCert configure -state readonly
		.saveCert.labExp.listCert configure -state readonly
#		puts $::listx509
	set nick [lindex $::listx509 0]
	set txt ".text.text"
	puts "NICK=$nick"
	if {[llength $::listx509]} {
	    
	    ::fileWithCert $txt $nick $::arrayCer($nick)
	} else {
	    $txt delete 0.0 end
	}
puts "UpdateListTok END"
#Автоматическое обновление токенов
#after 5000 updatetok
}

#######################
frame $w.tit -bg #ffffff
#puts "WW=$w"
# -highlightthickness 2 -highlightbackground #93cee9  -highlightcolor #93cee9
frame $w.tit.left -bg #ffffff -pady 0 
# -highlightthickness 2 -highlightbackground #93cee9  -highlightcolor #93cee9
#Combobox for tokens and certificate
frame .saveCert -background #93cee9
frame .selecttok -bg #ffffff
label .selecttok.lab  -text "[mc {Select token/smartcard}]:" -anchor w -font TkDefaultFontBold -width 0 -bg #ffffff -fg #000000
ttk::combobox .selecttok.tok -width 30 -textvariable nickTok -state readonly -values $::listtok -background #ffffff
bind .selecttok.tok <<ComboboxSelected>> {selectTok %W $nickTok }
button  .selecttok.update -command {updatetok} -image ::img::update_27x24 -compound left -bd 0 -background #ffffff -activebackground #ffffff -highlightthickness 0
pack .selecttok.lab  -side left -padx {0 5} -pady 2 -expand 0 -fill x
pack .selecttok.tok -side left -padx 0 -pady 2 -expand 1 -fill x -ipady 3
pack .selecttok.update -side right -padx {0 5} -pady 2 -expand 0 -fill none

frame .saveCert.labExp -borderwidth 2 -bg #ffffff -relief flat -highlightthickness 2 -highlightbackground #93cee9 -highlightcolor #93cee9

pack .saveCert.labExp -padx 0 -pady 0 -expand 0 -fill x
pack .selecttok  -in .saveCert.labExp -padx 0 -pady {3 0} -side top  -expand 1 -fill x
label .saveCert.labExp.labCert -image svitok_40x16 -compound left -text "[mc {Select certificate to view, import, export or remove}]:" -font TkDefaultFontBold -anchor w -bg #ffffff
pack .saveCert.labExp.labCert  -padx 0 -pady 0 -side top  -expand 1 -fill x
ttk::combobox .saveCert.labExp.listCert -width 50 -textvariable nickCert -state readonly -values $::listx509 -background #ffffff
bind .saveCert.labExp.listCert <<ComboboxSelected>> {selectCert %W $nickCert }

pack .saveCert.labExp.listCert -padx 4 -pady {6 4} -ipady 2  -expand 1 -fill x
frame .saveCert.labExp.but  -background snow
ttk::button .saveCert.labExp.but.butSave -command {set labCert [.saveCert.labExp.listCert get];
saveCertFile . .fr1.fr2_list $labCert} -text [mc "Export"] -style My.TButton 
pack .saveCert.labExp.but.butSave  -padx 4 -pady 4  -side left
ttk::button .saveCert.labExp.but.butCan  -command {global yespas;
set yespas "no" } -text [mc "Exit"] -style My.TButton 	-image ::img::delete -compound left -command exit

set bb [mc "Load certificate"]
set aa "ttk::button .saveCert.labExp.but.loadcert -text \"$bb\" -image ::img::view -compound left -style My.TButton -command {.text.text delete 0.0 end; ::fileWithCert  {.text.text} \"\" \"\"}"
#    puts $aa
set com [subst "$aa"]
eval $com
ttk::menubutton .saveCert.labExp.but.dopcom -text [mc "Additionaly"] -menu .saveCert.labExp.but.dopcom.menu -direction above  -style My.TButton
# MeB.TMenubutton
menu .saveCert.labExp.but.dopcom.menu -tearoff 0
.saveCert.labExp.but.dopcom.menu add command -label [mc "Load chain CA"] -command {::loadchain "chain"}
.saveCert.labExp.but.dopcom.menu add command -label [mc "Validaty by CRL"]  -command {::loadcrl}
set aa ".saveCert.labExp.but.dopcom.menu add command -label {[mc "Save certificate"]} -command {exportparsecert \".text.text\" }"
set com [subst "$aa"]
eval $com
.saveCert.labExp.but.dopcom.menu add command -label [mc "OCSP validity"]  -command {return}
.saveCert.labExp.but.dopcom.menu add command -label [mc "Verify signature"]  -command {::verifysign}
.saveCert.labExp.but.dopcom.menu add separator
.saveCert.labExp.but.dopcom.menu add command -label [mc "About"] -command {aboutUtil .about 0}

pack .saveCert.labExp.but.loadcert  -padx 4 -pady 4  -side left
pack .saveCert.labExp.but.dopcom  -padx 4 -pady 4  -side left
pack .saveCert.labExp.but.butCan  -padx 4 -pady 4  -side left


button  .saveCert.labExp.but.lang -command {changelang $w} -image ::img::us_32x32 -compound left -bd 0 -background #ffffff -activebackground #ffffff -highlightthickness 0
pack .saveCert.labExp.but.lang -in .saveCert.labExp.but -padx 4 -pady 4 -side right
pack .saveCert.labExp.but -in .saveCert.labExp -padx 4 -pady 4 -side bottom -expand 1 -fill x


frame .saveCert.labExp.typeCert -relief flat -bg #ffffff
pack .saveCert.labExp.typeCert -side top  -pady 3 -expand 1 -fill x
label .saveCert.labExp.typeCert.lab -text "[mc {Choose format to export}]:" -font TkDefaultFontBold -bg #ffffff -anchor w
pack .saveCert.labExp.typeCert.lab -side left -padx 5 -expand 0 -fill x  -anchor w
radiobutton .saveCert.labExp.typeCert.typeDer -anchor n -bg #ffffff -command {puts "DER=$formatCert"} -justify right -text [mc {DER-format}] -value 0 -variable formatCert -activebackground #93cee9 -bd 0  -highlightthickness 0
pack .saveCert.labExp.typeCert.typeDer -expand 0 -fill x  -side left -padx 5 
radiobutton .saveCert.labExp.typeCert.typePem -anchor center -bg #ffffff -borderwidth 3 -command {puts "PEM=$formatCert"} -justify left -text [mc {PEM-format}] -value 1 -variable formatCert -activebackground #93cee9 -bd 0  -highlightthickness 0
pack .saveCert.labExp.typeCert.typePem -expand 0 -fill x -side left

frame .saveCert.labExp.typeSep  -height 3 -bd 2 -relief groove -bg #e2e2e1
pack .saveCert.labExp.typeSep   -pady 0  -padx 0 -expand 1 -fill x -side top

pack .saveCert -in $w.tit.left -padx 0 -pady 0 -anchor n -expand 1 -fill x

##########################################

option add *Label.background #ffffff
option add *Label.highlightbackground #ffffff
#ttk::button -shiftrelief 0 - кнопка не углубляется
ttk::style configure Me.TButton -background #ffffff
ttk::style configure Me.TButton  -bd 0  -anchor sw  -shiftrelief 0 -style Me.TButton  -font TkDefaultFontBold \
     -relief flat -bg #ffffff -highlightbackground #ffffff -activebackground #ffffff
ttk::style map Me.TButton -background [list disabled #ffffff pressed #ffffff  active #ffffff] -foreground [list disabled #a3a3a3] -relief [list {pressed !disabled} flat]
ttk::button $w.tit.left.caption -text [mc "Details"] -image ::img::not_det  -compound left  -style Me.TButton  -command {typeview $w}
#button $w.tit.left.caption -text [mc "Details"] -font TkDefaultFontBold -bd 0  -anchor sw -image ::img::not_det  \
#     -relief flat -compound left -command {typeview $w} -bg #ffffff -highlightbackground #ffffff -activebackground #ffffff

pack $w.tit.left.caption -side top -expand yes -pady 0 -anchor w -padx 5 

pack $w.tit -side top -fill both -expand 1
pack $w.tit.left  -side left -expand 1 -padx 0 -pady {0 0} -fill both 

##################
text $w.text -yscrollcommand [list $w.scroll set] -setgrid 1  -relief flat -font TkDefaultFont -exportselection 1 -wrap word \
	-height 20 -undo 1 -autosep 1 -bg #ffffff  -highlightbackground #ffffff	 -highlightcolor #ffffff 

switch $typesys {
    win32        { 
ttk::scrollbar $w.scroll -command [list $w.text yview] 
    }
    x11 { 
	ttk::style map My.Vertical.TScrollbar -troughcolor [list disabled wheat pressed #a3a3a3  active #c3c3c3] -foreground [list disabled #a3a3a3] -relief [list {pressed !disabled} sunken] -background  [list disabled #93cee9 pressed #93cee9  active #93cee9] 
	ttk::style map My.Vertival.TScrollbar -arrowcolor [list disabled #ffffff pressed #ffffff  active #ffffff ]
	ttk::style configure My.Vertical.TScrollbar -troughcolor {#eff0f1} 
	ttk::style configure My.Vertical.TScrollbar -arrowcolor {#eff0f1}
	ttk::style configure My.Vertical.TScrollbar -background #898989 -width 0 -relief flat -borderwidth 0 -arrowsize 5 -arrowcolor #ffffff
	ttk::scrollbar $w.scroll -command [list $w.text yview] -orient vertical -style My.Vertical.TScrollbar
	.saveCert.labExp.but.butSave configure -style My.TButton
	.saveCert.labExp.but.butCan configure -style My.TButton
    }
    classic - aqua { 
ttk::scrollbar $w.scroll -command [list $w.text yview] 
    }
}


bind $w.text <ButtonPress-3> {showContextMenu %W %x %y %X %Y}

#pack $w.text1 -expand yes -fill both -side bottom
$w.text configure -foreground #ffffff
#pack $w.scroll -side right -fill y
pack $w.scroll -anchor center -expand 0 -fill y -side right  -padx {0 5} -pady {2 8}

pack $w.text -expand yes -fill both -side bottom
pack forget $w.scroll
#SEPARATOR
#    frame $w.sep -height 1 -relief flat -bg #b6b6b3
#pack $w.sep -side bottom -fill x -expand 1 -pady 0
$w.text tag configure bold -font TkDefaultFontBold
#font configure TkDefaultFont -size 12
$w.text tag configure super -offset 4p -font TkDefaultFont
$w.text tag configure sub -offset -2p -font TkDefaultFont
$w.text tag configure margins -lmargin1 4m -lmargin2 12m -rmargin 10m
$w.text tag configure margins1 -lmargin1 2m -lmargin2 2.2i -rmargin 2m -spacing1 1p -spacing2 1p -spacing3 1p
$w.text tag configure margins11 -lmargin1 2m -lmargin2 2.2i -rmargin 2m -spacing1 1p -spacing2 1p -spacing3 1p -font TkFixedFont
$w.text tag configure margins2 -lmargin1 6m -lmargin2 2.2i -rmargin 2m -spacing1 1p -spacing2 1p -spacing3 1p
$w.text tag configure spacing -spacing1 10p -spacing2 2p \
	-lmargin1 12m -lmargin2 6m -rmargin 10m
$w.text delete 0.0 end
$w.text configure -tabs {1m 2.2i}
if {$lang == "ru"} {
    .saveCert.labExp.but.lang configure -image  ::img::ru_32x32 -compound left
} else {
    .saveCert.labExp.but.lang configure -image  ::img::us_32x32 -compound left
}
if {$loadfile != ""} {
    fileWithCert $w $loadfile
}
typeview $w

toplevel .topPinPw -borderwidth 2 -class Toplevel -relief ridge -background #39b5da -highlightbackground #d9d9d9 -highlightcolor black 
global icon11_32x32
wm geometry .topPinPw +383+258
wm iconphoto .topPinPw  ::img::logo
wm state .topPinPw withdraw
labelframe .topPinPw.labFrPw -borderwidth 4 -class LabelFrame -labelanchor nw -relief groove -text "[mc {Enter the PIN code and press Enter}]" -foreground blue -background #eff0f1 
pack .topPinPw.labFrPw -anchor nw -padx 3 -pady 3 -fill x -expand 1
entry .topPinPw.labFrPw.entryPw -background snow -highlightbackground #d9d9d9 -highlightcolor blue -justify left -relief sunken -show * -width 40
entry .topPinPw.labFrPw.entryLb -background snow -highlightbackground #d9d9d9 -highlightcolor blue -justify left -relief sunken -width 40
pack .topPinPw.labFrPw.entryPw -fill x -expand 1 -padx 5 -pady 5  -ipady 2 
bind .topPinPw.labFrPw.entryPw <Key-Return> {readPw .topPinPw.labFrPw.entryPw}
bind .topPinPw.labFrPw.entryLb <Key-Return> {readPw .topPinPw.labFrPw.entryLb}
ttk::button .topPinPw.labFrPw.butPw  -command {global yespas;set yespas "no"; wm state .topPinPw withdraw;} -text [mc "Cancel"]  -style My.TButton
#pack .topPinPw.labFrPw.butPw -pady 5 -sid right -padx 5
pack .topPinPw.labFrPw.butPw -pady 5 -sid bottom -padx 5 -anchor e

